"""Globally defined subsystem IDs.

All error codes returns from an IOTile device are in either short or long
format.

Short form error codes are 16-bit values that are declared in the errors
module.

Long form error codes have a 16-bit value and a 16-bit subsystem ID.  This
module declares the globally known subsystems.

For both error codes and subsystem IDs, values < 0x8000 are given global
meaning independent of any tile.  Values >= 0x8000 are private to the given
tile that is returning them and must be interpretted in the context of that
tile.
"""

from enum import IntEnum

class ControllerSubsystem(IntEnum):
    """All core subsystems defined for an IOTile controller."""

    TILE_MANAGER = 0x8000
    """The tile management subsystem."""

    PERSISTENT_STORAGE = 0x8001
    """Low level routines dealing with persistent storage."""

    SENSOR_LOG = 0x8002
    """The sensor log subsystem."""

    SENSOR_GRAPH = 0x8003
    """The sensor graph engine subsystem."""

    CONTROLLER = 0x8004
    """Generic controller related routines."""

    REMOTE_BRIDGE = 0x8005
    """The remote bridge subsystem."""

    SECURITY = 0x8006
    """The security subsystem."""



"""Mixin for device updating via signed scripts."""

import base64
from iotile.core.hw.virtual import tile_rpc
from iotile.core.hw.update import UpdateScript
from ...virtual import SerializableState
from .controller_system import ControllerSubsystemBase


# FIXME: Move this to constants section
class BRIDGE_STATUS(object):
    """Enum for valid remote bridge statuses."""
    IDLE = 0
    WAITING = 1
    RECEIVING = 2
    RECEIVED = 3
    VALIDATED = 4
    EXECUTING = 5


class RemoteBridgeState(SerializableState, ControllerSubsystemBase):
    """Serializeable state object for all remote bridge state.

    Note that the script_error property is just a convenience property
    for looking at internal exceptions when executing scripts.  It does
    not reflect a real emulated device state and is not dumped or restored
    when dump() or restore() is called.
    """

    def __init__(self, emulator):
        ControllerSubsystemBase.__init__(self, emulator)
        SerializableState.__init__(self)

        self.status = BRIDGE_STATUS.IDLE
        self.error = 0
        self.parsed_script = None
        self.script_error = None

        self.mark_ignored('initialized')
        self.mark_complex('parsed_script', self._dump_script, self._restore_script)

    def clear_to_reset(self, config_vars):
        """Clear the RemoteBridge subsystem to its reset state."""

        super(RemoteBridgeState, self).clear_to_reset(config_vars)
        self.status = BRIDGE_STATUS.IDLE
        self.error = 0

    def _dump_script(self, value):
        if value is None:
            return None

        encoded = value.encode()
        return base64.b64encode(encoded).decode('utf-8')

    @classmethod
    def _restore_script(cls, b64encoded):
        if b64encoded is None:
            return None

        encoded = base64.b64decode(b64encoded)
        return UpdateScript.FromBinary(encoded)


class RemoteBridgeMixin(object):
    """Reference controller subsystem for device updating.

    This class must be used as a mixin with a ReferenceController base class.
    """


    def __init__(self, emulator):
        self.remote_bridge = RemoteBridgeState(emulator)
        self._post_config_subsystems.append(self.remote_bridge)

    @tile_rpc(0x2100, "", "L")
    def begin_script(self):
        """Indicate we are going to start loading a script."""

        if self.remote_bridge.status in (BRIDGE_STATUS.RECEIVED, BRIDGE_STATUS.VALIDATED, BRIDGE_STATUS.EXECUTING):
            return [1]  #FIXME: Return correct error here

        self.remote_bridge.status = BRIDGE_STATUS.WAITING
        self.remote_bridge.error = 0
        self.remote_bridge.script_error = None
        self.remote_bridge.parsed_script = None

        self._device.script = bytearray()

        return [0]

    @tile_rpc(0x2102, "", "L")
    def end_script(self):
        """Indicate that we have finished receiving a script."""

        if self.remote_bridge.status not in (BRIDGE_STATUS.RECEIVED, BRIDGE_STATUS.WAITING):
            return [1] #FIXME: State change

        self.remote_bridge.status = BRIDGE_STATUS.RECEIVED
        return [0]

    @tile_rpc(0x2103, "", "L")
    def trigger_script(self):
        """Actually process a script."""

        if self.remote_bridge.status not in (BRIDGE_STATUS.RECEIVED,):
            return [1] #FIXME: State change

        # This is asynchronous in real life so just cache the error
        try:
            self.remote_bridge.parsed_script = UpdateScript.FromBinary(self._device.script)
            #FIXME: Actually run the script

            self.remote_bridge.status = BRIDGE_STATUS.IDLE
        except Exception as exc:
            self._logger.exception("Error parsing script streamed to device")
            self.remote_bridge.script_error = exc
            self.remote_bridge.error = 1 # FIXME: Error code

        return [0]

    @tile_rpc(0x2104, "", "LL")
    def query_status(self):
        """Get the status and last error."""

        return [self.remote_bridge.status, self.remote_bridge.error]

    @tile_rpc(0x2105, "", "L")
    def reset_script(self):
        """Clear any partially received script."""

        self.remote_bridge.status = BRIDGE_STATUS.IDLE
        self.remote_bridge.error = 0
        self.remote_bridge.parsed_script = None
        self._device.script = bytearray()

        return [0]



"""The canonical reference implementations of a generic IOTile device and a minimal IOTile controller tile."""

from .reference_device import ReferenceDevice
from .reference_controller import ReferenceController

__all__ = ['ReferenceDevice', 'ReferenceController']



"""Utility function to pretty print an rpc call and response."""

from binascii import hexlify
from ..constants import rpc_name

def format_rpc(data):
    """Format an RPC call and response.

    Args:
        data (tuple): A tuple containing the address, rpc_id, argument and
            response payloads and any error code.

    Returns:
        str: The formated RPC string.
    """

    address, rpc_id, args, resp, _status = data

    name = rpc_name(rpc_id)

    if isinstance(args, (bytes, bytearray)):
        arg_str = hexlify(args)
    else:
        arg_str = repr(args)

    if isinstance(resp, (bytes, bytearray)):
        resp_str = hexlify(resp)
    else:
        resp_str = repr(resp)

    #FIXME: Check and print status as well
    return "%s called on address %d, payload=%s, response=%s" % (name, address, arg_str, resp_str)



"""Test coverage of the ReferenceDevice and ReferenceController emulated objects."""

import sys
import pytest
from iotile.core.hw import HardwareManager
from iotile.core.exceptions import HardwareError
from iotile.core.hw.proxy.external_proxy import find_proxy_plugin
from iotile.emulate.virtual import EmulatedPeripheralTile
from iotile.emulate.reference import ReferenceDevice
from iotile.emulate.constants import rpcs, Error
from iotile.emulate.transport import EmulatedDeviceAdapter


@pytest.fixture(scope="function")
def reference():
    """Get a reference device with a controller and single peripheral tile."""

    device = ReferenceDevice({'simulate_time': False})
    peripheral = EmulatedPeripheralTile(10, device)
    device.add_tile(10, peripheral)

    device.start()
    yield device, peripheral
    device.stop()


@pytest.fixture(scope="function")
def reference_hw():
    """Get a reference device and connected HardwareManager."""

    device = ReferenceDevice({'simulate_time': False})
    peripheral = EmulatedPeripheralTile(11, device)
    peripheral.declare_config_variable("test 1", 0x8000, 'uint16_t')
    peripheral.declare_config_variable('test 2', 0x8001, 'uint32_t[5]')

    device.add_tile(11, peripheral)

    adapter = EmulatedDeviceAdapter(None, devices=[device])

    with HardwareManager(adapter=adapter) as hw:
        hw.connect(1)

        yield hw, device, peripheral


def test_basic_usage():
    """Make sure we can import and use the objects."""

    with HardwareManager(port='emulated:reference_1_0@#eyJzaW11bGF0ZV90aW1lIjogZmFsc2V9') as hw:
        hw.connect(1)
        debug = hw.debug()

        state = debug.dump_snapshot()
        debug.restore_snapshot(state)


def test_peripheral_tiles():
    """Make sure the controller tile properly brings up the peripheral tiles in a controlled manner."""

    # Don't use the fixture since the purpose of this test is to make sure the fixture works
    device = ReferenceDevice({'simulate_time': False})
    peripheral = EmulatedPeripheralTile(10, device)
    device.add_tile(10, peripheral)

    device.start()
    device.stop()


def test_config_variable_rpcs(reference):
    """Make sure that we can set, list and get config variables."""

    device, peripheral = reference

    peripheral.declare_config_variable("test 1", 0x8000, 'uint16_t')
    peripheral.declare_config_variable('test 2', 0x8001, 'uint32_t[5]')

    # Test listing
    resp = device.rpc(10, rpcs.LIST_CONFIG_VARIABLES, 0)
    count = resp[0]
    ids = resp[1:]

    assert len(ids) == 9
    assert count == 2
    assert ids[:2] == (0x8000, 0x8001)

    # Test describing
    resp1 = device.rpc(10, rpcs.DESCRIBE_CONFIG_VARIABLE, 0x8000)
    resp2 = device.rpc(10, rpcs.DESCRIBE_CONFIG_VARIABLE, 0x8001)
    resp3 = device.rpc(10, rpcs.DESCRIBE_CONFIG_VARIABLE, 0x8002)

    assert len(resp1) == 5
    assert len(resp2) == 5
    assert len(resp3) == 5
    assert resp3[0] == Error.INVALID_ARRAY_KEY

    err, _, _, config_id, packed_size = resp1
    assert err == 0
    assert config_id == 0x8000
    assert packed_size == 2

    err, _, _, config_id, packed_size = resp2
    assert err == 0
    assert config_id == 0x8001
    assert packed_size == (1 << 15) | (5*4)

    # Test setting (make sure to artificially force pre-app started state)
    peripheral.initialized.clear()
    err, = device.rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8000, 0, bytes(bytearray([5, 6])))
    assert err == 0

    err, = device.rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8001, 0, bytes(bytearray([7, 8])))
    assert err == 0

    err, = device.rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8001, 19, bytes(bytearray([7, 8])))
    assert err == Error.INPUT_BUFFER_TOO_LONG

    err, = device.rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8002, 0, bytes(bytearray([7, 8])))
    assert err == Error.INVALID_ARRAY_KEY

    peripheral.initialized.set()
    err, = device.rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8000, 0, bytes(bytearray([5, 6])))
    assert err == Error.STATE_CHANGE_AT_INVALID_TIME

    # Test Getting
    data, = device.rpc(10, rpcs.GET_CONFIG_VARIABLE, 0x8000, 0)
    assert data == bytearray([5, 6])

    data, = device.rpc(10, rpcs.GET_CONFIG_VARIABLE, 0x8001, 0)
    assert data == bytearray([7, 8])

    data, = device.rpc(10, rpcs.GET_CONFIG_VARIABLE, 0x8002, 0)
    assert data == bytearray([])


def test_snapshot_save_load(reference):
    """Make sure snapshot saving and loading works without error on a peripheral tile."""

    device, _peripheral = reference

    state = device.dump_state()
    device.restore_state(state)


def test_config_database(reference_hw):
    """Test the controller config database RPCs using the canonical proxy plugin."""

    hw, _device, _peripheral = reference_hw

    con = hw.get(8, basic=True)
    slot = hw.get(11, basic=True)
    config_db = find_proxy_plugin('iotile_standard_library/lib_controller', 'ConfigDatabasePlugin')(con)

    # Make sure all of the basic RPCs work
    assert config_db.count_variables() == 0
    config_db.clear_variables()

    # Make sure we can set and get config variables
    config_db.set_variable('slot 1', 0x8000, 'uint16_t', 15)
    config_db.set_variable('slot 1', 0x8001, 'uint32_t[]', "[5, 6, 7]")

    assert config_db.count_variables() == 2
    var8000 = config_db.get_variable(1)
    var8001 = config_db.get_variable(2)

    var8000['metadata'] = str(var8000['metadata'])
    var8001['metadata'] = str(var8001['metadata'])

    assert var8000 == {
        'metadata': 'Target: slot 1\nData Offset: 0\nData Length: 4\nValid: True',
        'name': 0x8000,
        'data': bytearray([15, 0])
    }
    assert var8001 == {
        'metadata': 'Target: slot 1\nData Offset: 4\nData Length: 14\nValid: True',
        'name': 0x8001,
        'data': bytearray([5, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0])
    }

    slot_conf = slot.config_manager()

    # Make sure we stream all variables on a reset
    slot.reset(wait=0)
    assert slot_conf.get_variable(0x8000) == bytearray([15, 0])
    assert slot_conf.get_variable(0x8001) == bytearray([5, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0])

    # Make sure invalidated variables are not streamed and the tile resets them on reset
    config_db.invalidate_variable(1)
    assert config_db.count_variables() == 2
    slot.reset(wait=0)
    assert slot_conf.get_variable(0x8000) == bytearray()
    assert slot_conf.get_variable(0x8001) == bytearray([5, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0])

    # Check memory limits before and after compaction
    usage = config_db.memory_limits()
    assert usage == {
        'data_usage': 18,
        'data_compactable': 4,
        'data_limit': 4096,
        'entry_usage': 2,
        'entry_compactable': 1,
        'entry_limit': 255
    }

    # Make sure we remove invalid and keep valid entries upon compaction
    config_db.compact_database()
    usage = config_db.memory_limits()
    assert usage == {
        'data_usage': 14,
        'data_compactable': 0,
        'data_limit': 4096,
        'entry_usage': 1,
        'entry_compactable': 0,
        'entry_limit': 255
    }

    assert config_db.count_variables() == 1
    slot.reset(wait=0)
    assert slot_conf.get_variable(0x8000) == bytearray()
    assert slot_conf.get_variable(0x8001) == bytearray([5, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0])


def test_tile_manager(reference_hw):
    """Test the controller tile_manager RPCs using the canonical proxy plugin."""

    # Work around inconsistency in output of iotile-support-lib-controller-3 on python 2 and 3
    if sys.version_info.major < 3:
        con_str = 'refcn1, version 1.0.0 at slot 0'
        peri_str = 'noname, version 1.0.0 at slot 1'
    else:
        con_str = "b'refcn1', version 1.0.0 at slot 0"
        peri_str = "b'noname', version 1.0.0 at slot 1"

    hw, _device, _peripheral = reference_hw

    con = hw.get(8, basic=True)
    slot = hw.get(11, basic=True)
    tileman = find_proxy_plugin('iotile_standard_library/lib_controller', 'TileManagerPlugin')(con)

    assert tileman.count_tiles() == 2

    con = tileman.describe_tile(0)
    peri = tileman.describe_tile(1)
    assert str(con) == con_str
    assert str(peri) == peri_str

    # Make sure reseting a tile doesn't add a new entry
    slot.reset(wait=0)

    assert tileman.count_tiles() == 2

    con = tileman.describe_tile(0)
    peri = tileman.describe_tile(1)
    assert str(con) == con_str
    assert str(peri) == peri_str

    con = tileman.describe_selector('controller')
    peri = tileman.describe_selector('slot 1')
    assert str(con) == con_str
    assert str(peri) == peri_str


def test_raw_sensor_log(reference_hw):
    """Test to ensure that the raw sensor log works."""

    hw, _device, _peripheral = reference_hw

    con = hw.get(8, basic=True)
    sensor_graph = find_proxy_plugin('iotile_standard_library/lib_controller', 'SensorGraphPlugin')(con)

    assert sensor_graph.count_readings() == {
        'streaming': 0,
        'storage': 0
    }

    assert sensor_graph.highest_id() == 0

    sensor_graph.push_many('output 1', 10, 20)
    sensor_graph.push_many('buffered 1', 15, 25)

    assert sensor_graph.highest_id() == 45

    readings = sensor_graph.download_stream('output 1')
    assert len(readings) == 20

    readings = sensor_graph.download_stream('buffered 1')
    assert len(readings) == 25

    readings = sensor_graph.download_stream('buffered 1', reading_id=44)
    assert len(readings) == 2

    sensor_graph.clear()
    assert sensor_graph.count_readings() == {
        'streaming': 1,
        'storage': 0
    }

    assert sensor_graph.highest_id() == 46


def test_rsl_reset_config(reference_hw):
    """Make sure we properly load config variables into the sensor_log."""

    hw, _device, _peripheral = reference_hw

    con = hw.get(8, basic=True)
    sensor_graph = find_proxy_plugin('iotile_standard_library/lib_controller', 'SensorGraphPlugin')(con)
    config = find_proxy_plugin('iotile_standard_library/lib_controller', 'ConfigDatabasePlugin')(con)

    config.set_variable('controller', 0x2004, 'uint8_t', 1)
    con.reset(wait=0)

    with pytest.raises(HardwareError):
        sensor_graph.push_many('buffered 1', 15, 20000)

    assert sensor_graph.count_readings() == {
        'streaming': 0,
        'storage': 16128
    }

    config.set_variable('controller', 0x2005, 'uint8_t', 1)
    con.reset(wait=0)

    with pytest.raises(HardwareError):
        sensor_graph.push_many('output 1', 15, 50000)

    assert sensor_graph.count_readings() == {
        'streaming': 48896,
        'storage': 16128
    }


def test_rsl_dump_restore(reference_hw):
    """Make sure the rsl state is properly saved and restored."""

    hw, device, _peripheral = reference_hw
    refcon = device.controller

    con = hw.get(8, basic=True)
    sensor_graph = find_proxy_plugin('iotile_standard_library/lib_controller', 'SensorGraphPlugin')(con)

    # Verify that we properly restore data and our download walker
    for i in range(0, 10):
        sensor_graph.push_reading('output 1', i)

    con.rpc(0x20, 0x08, 0x5001, result_format="LLLL")
    err, _timestamp, reading, unique_id, _act_stream = con.rpc(0x20, 0x09, 1, arg_format='B', result_format="LLLLH2x")
    assert err == 0
    assert reading == 0
    assert unique_id == 1

    state = device.dump_state()
    device.restore_state(state)

    # Make sure we keep our same place in the dump stream line
    err, _timestamp, reading, unique_id, _act_stream = con.rpc(0x20, 0x09, 1, arg_format='B', result_format="LLLLH2x")

    assert err == 0
    assert reading == 1
    assert unique_id == 2

    # Make sure we have all of our data
    assert refcon.sensor_log.engine.count() == (0, 10)

    readings = sensor_graph.download_stream('output 1')
    assert len(readings) == 10
    for i, reading in enumerate(readings):
        assert reading.value == i
        assert reading.reading_id == i + 1



r"""RPCs implemented by the SensorGraph subsystem on an IOTile controller.

The SensorGraph subsystem is designed to allow you to embed a small set
automatic actions that should be run whenever events happen inside the device.
Sensorgraph is structured as a dependency tree where actions are linked by
named FIFOs that route data from inputs -> processing functions -> output
sinks.

The sensor-graph language and implementation were designed to facilitate
static analysis of maximum resource needs as well as runtime properties like
power consumption.

The sensor-graph subsystem interacts with the sensor-log and stream-manager
subsystems in the following way:

sensor-graph has rules that respond to events on a device and generate data
in named FIFOs called streams.  All of the actual data storage and stream
management on behalf of sensor-graph is handled by the sensor-log subsystem.

The streaming subsystem listens on configured streams to see if it should
package them up and send them to an external user in the form of a `report`.

So the interaction diagram looks like:


    sensor-graph       <------XXXX------>   stream-manager
                       no direct contact

    generates data in streams               sole purpose in life is to
    according to configured rules           build reports from streams

                    \\                  //
      producer       \\                //   consumer only
      and consumer    \\              //    (generates no readings!)
                          sensor-log

                          stores all streams and lets you
                          inspect their contents using
                          stream walkers that automatically
                          get updated when new data is available.

The actual work of simulating the functionality of the embedded sensor-graph
engine is performed by the iotile-sensorgraph package that makes it available
in a variety of contexts including a direct command line simulator named
iotile-sgrun as well as an optimizing compiler.
"""

from .rpc_declaration import RPCDeclaration


SG_COUNT_NODES = RPCDeclaration(0x2002, "", "L")
"""Count the number of nodes currently in the sensor graph.

This RPC just returns the number of nodes currently added
to the sensor graph.  It can be called at any time and
cannot fail.

Returns:
  - uint32_t: The number of nodes currently present.
"""

SG_ADD_NODE = RPCDeclaration(0x2003, "20s", "L")
"""Add a node to the current sensor graph.

This RPC is designed to be called repeatedly to build up a sensor graph. You
call it with a complicated binary descriptor object that describes the node
you want to build and the sensor-graph engine adds that node to its internal
graph structure.  There are a number of potential error codes that can be
returned if there is an inssue with your node descriptor.

This is a synchronous RPC, so when it returns the node has already been added
to the graph and it is ready for usage.

It is safe to call this RPC at any point, even when the sensor-graph engine
is enabled.

You can see the structure of a node descriptor and convert it to/from a nice
human-readable string using routines from `iotile.sg.node_descriptor`:

  - parse_binary_node_descriptor()
  - create_binary_descriptor()

Args:
  - char[20]: A complex serialized binary node descriptor.

Returns:
  - uint32_t: An error code.

Possible Errors:
  - INVALID_NODE_STREAM: You cannot create a node with an output stream of
    type 'input'. 'input' streams are global sensor-graph inputs so they
    cannot be generated as the result of node.  They are created using the
    SG_GRAPH_INPUT RPC.
  - NO_NODE_SPACE_AVAILABLE: You have run out of space for new graph nodes.
  - NO_AVAILABLE_OUTPUTS: Each node has a preallocated array linking it to any
    possible output nodes that should be triggered it produces a new reading.
    If you have a node that fans out too quickly, you may get this error.  It
    can be fixed by inserting another node to buffer the output.
"""

SG_GRAPH_INPUT = RPCDeclaration(0x2004, "LH", "L")
"""Present a new input to the sensor graph.

This is the only method you can call to trigger the sensor-graph engine to
process a new reading.  While you can directly push readings to the underly
rsl streams using RSL_PUSH, that method is not guaranteed to fire relevant
sensor-graph rules and must be used with extreme caution when there is a
loaded sensor-graph.

You specify the stream that you want to push to, which should be an input
class stream and the value that you want to push.  It will automatically be
timestamped with the current utc time or up-time of the controller.

**This is an asynchronous RPC.**

Unlike most sensor-graph related RPCs, this RPC will return immediately after
queueing the input and not wait until it is fully processed.  This behavior is
critical since the processing of the input may require calling an RPC because
it triggered a node that has a `call_rpc` processing function.  If we were
waiting synchronously for processing to be finished this would deadlock since
the RPC bus would be blocked so the `call_rpc` function would hang and we
would wait forever.

Args:
  - uint32_t: The reading value that you wish to present to the sensor-graph
    engine.  If no nodes are configured to listen to this input, it will be
    ignored silently.
  - uint16_t: The stream id of the **input stream** that you want to push.
    This must be an **input** stream.

Returns:
  - uint32_t: An error code.  There are currently no errors possible.
"""


SG_SET_ONLINE = RPCDeclaration(0x2005, "H", "L")
"""Pause or resume the sensor-graph engine.

If you call SG_SET_ONLINE with a false value, it will pause the processing
or all future inputs until you call SG_SET_ONLINE with a true value.

Pausing means that any call to SG_GRAPH_INPUT will be dropped.  All other
functionality still works the same whether the graph is paused or resumed.

It is always safe to gratuitously call this RPC with the same state.  This is
good because there is no way to determine whether the sensor-graph is
currently paused or resumed.

Args:
  - uint16_t: A true value to enable the sensor-graph and a false value to disable it.

Returns:
  - uint32_t: An error code.  Currently, there are no possible errors that could be
    returned.
"""

SG_ADD_STREAMER = RPCDeclaration(0x2007, "V", "L")
"""Add a streamer to the sensor-graph.

A streamer is a resource inside the sensor-graph subsystem that packages up
readings in one or more streams (that you select with a selector programmed as
part of the streamer resource).  Whenever given conditions are met, the
streamer will package up all of its readings into a report that it sends out
of the device using the device's streaming interface.

You specify how the streamer should be configured by using a binary streamer
descriptor similar to how you specify a node using a node descriptor structure.

You can see the descriptor structure and utility functions in the module:
iotile.sg.streamer

The streamers are added sequentially every time you call SG_ADD_STREAMER so
the first time you call this RPC, the streamer will get index 0, the next
call will be index 1, etc.

Args:
  - char[14]: A binary streamer descriptor.  The final byte is padding though
    so you can pass either 13 or 14 bytes of data.

Returns:
  - uint32_t: an error code.

Possible Errors:
  - NO_MORE_STREAMER_RESOURCES: There are no more available streamer
    resources.
"""

#FIXME: Move streamer info structure to iotile.sg
SG_QUERY_STREAMER = RPCDeclaration(0x200a, "H", "V")
"""Query the current status of a streamer resource by its index.

This method can be called at any time to inspect the current status of a
streamer. You can learn things like when the last time it was triggered was,
what error code was returned the last time it was triggered and whether it is
currently in the process of creating and/or streaming a report.

Polling this RPC is the correct way to determine what a streamer is doing.

Args:
  - uint16_t: The index of the streamer that you wish to query.

Returns (option 1):
  - char[20]: A binary streamer status structure.

Returns (option 2):
  - uint32_t: An error code in case the index is invalid.

Possible Errors:
  - STREAMER_NOT_ALLOCATED: If the index corresponds to a streamer that has
    not been allocated.
"""

SG_RESET_GRAPH = RPCDeclaration(0x200d, "", "L")
"""Clear the current sensor-graph in ram and also any persisted graph.

This function completely clears any stored sensor-graph nodes or streamers. It
does not impact any data stored in the raw sensor log.

There is no way to clear just the persisted sensor-graph or the one in ram
without also clearing the persisted one.

This function will also clear the persistent storagea of any received streamer
acknowledgement values.  See SG_SEEK_STREAMER for a discussion of what these
values are used for.

Returns:
  - uint32_t: An error code.  No errors are currently possible.
"""

SG_PERSIST_GRAPH = RPCDeclaration(0x200e, "", "L")
"""Save the sensor-graph currently in memory to persistent storage.

The normal way you build up a sensor-graph is by repeatedly calling
SG_ADD_NODE and SG_ADD_STREAMER until you have the desired graph
structure in memory.

Then you call SG_PERSIST_GRAPH to save that to persistent storage. Once you
call SG_PERSIST_GRAPH, the sensor-graph will automatically be loaded and start
running every time the device resets so this is also the way you "launch" the
device onto its autonomous mission.

This is a synchronous RPC, so it does not return until the graph is completely
persisted.

Returns:
  - uint32_t: An error code.  No errors are currently possible.
"""

SG_SEEK_STREAMER = RPCDeclaration(0x200f, "HHL", "L")
"""Inform the given streamer about what readings have been successfully received.

Streamer resources are designed to act like reliable TCP streams that never
throw away data until it is positively acknowledged as received by the remote
party.  Since the streaming interface is unidirectional, there needs to be a
way for that remote party to inform the sensor-graph subsystem that it has
received all readings up to sequence number X.

That is the purpose of this seek streamer RPC.  It lets a remote party
gratuitious inform the IOTile device what the highest reading_id is has
received from each streamer.  This is used to advance an internal pointer in
each streamer resource to not send any of that data again.

Since these acknowledgement values may be received via intermediaries that may
have out of date information, by default the acknowledgement value can only
ever increase (which should be the case in reality).  This makes it safe for
everyone to just gratuitously send their highest acknowledgement value
whenever they get a chance and the IOTile device will ignore old values.

If you need to roll a streamer back to fix an error where you acknowledged a
reading but then later lost it, you can set the force parameter in this RPC to
true and it will accept whatever acknowledgement value you set, even if it is
lower than what it currently has stored.

The behavior of this RPC when an incorrect acknowledgement value is given
can be counterintuitive so it's important to understand what will happend when
each of the following error codes are returned:

  - STREAMER_NOT_ALLOCATED: No change
  - OLD_ACKNOWLEDGE_UPDATE: No change
  - ID_FOUND_FOR_ANOTHER_STREAM: This is a protocol error, so the streamer attempts
    to get back in sync with the remote party by pointing itself at the oldest
    selected reading.
  - NO_MORE_READINGS: The streamer will skip all readings and point at the end
    of the storage area.

This specific behavior is not to considered stable, documented behavior since
it is undesireable in many cases.  A future variant of the iotile controller
may change how streamers react to protocol errors.

Args:
  - uint16_t: The index of the streamer you want to acknowledge.
  - uint16_t: true to force a rollback, false to ignroe the acknowledgement if
    it would roll this streamer backward in time.
  - uint32_t: The highest reading_id that the RPC caller is claiming to have
    received.  The streamer will no longer send readings in reports that have
    reading ids <= this value.

Returns:
  - uint32_t: an error code.

Possible Errors:
  - STREAMER_NOT_ALLOCATED: If the index corresponds to a streamer that has
    not been allocated.
  - OLD_ACKNOWLEDGE_UPDATE: The acknowledgement value is smaller than what is
    currently stored for this streamer and force was false.  If you need to
    roll the streamer back you need to pass the force parameter as true.
  - ID_FOUND_FOR_ANOTHER_STREAM: The reading_id passed corresponds to a reading
    that was never selected by this streamer so it cannot have been received.
  - NO_MORE_READINGS: The reading_id passed is greater than all reading ids
    stored in this iotile device to it cannot possibly have actually have
    been received.
"""

SG_TRIGGER_STREAMER = RPCDeclaration(0x2010, "H", "L")
"""Manually trigger a stream to attempt to create/stream a report immediately.

This function can be used to test out a streamer or as part of an action
programmed into your sensor-graph script to control when the device attempts
to send data remotely.

Just because you call this function on a valid streamer does not mean it will
necessarily generate a report.  It also needs to have unacknowledged data that
needs to be streamed.  You can inspect the return value of this RPC to see
whether a report was indeed queued for streaming.

Args:
  - uint16_t: The index of the streamer that you want to trigger.

Returns:
  - uint32_t: An error code.

Possible Errors:
  - STREAMER_NOT_ALLOCATED: If the index corresponds to a streamer that has
    not been allocated.
  - STREAM_ALREADY_IN_PROGRESS: The streamer is currently creating a report
    so you cannot trigger it again until it finishes and returns to the
    idle state.
  - STREAMER_HAS_NO_NEW_DATA: The streamer does not have any unacknowledged
    readings to send so the report would be empty.  No report is generated
    in this case.
  - INVALID_STREAM_DESTINATION: The configured tile that this streamer is
    supposed to send its report to does not exist.  This is usually a fatal
    configuration error telling you that your streamer is misconfigured.
"""

SG_INSPECT_GRAPH_NODE = RPCDeclaration(0x2016, "H", "20s")
"""Get the binary node descriptor from a node by its index.

This allows you to verify the action of the SG_ADD_NODE RPC by inspecting the
last node added to the sensor-graph.  The binary descriptor returned should
be identical to what you used to create it.

This RPC is also useful for inspecting what sensor-graph is programmed into a
given device by iterating over all nodes and returning their node descriptors.

**If the given index is not in use, this RPC will not complete normally but
instead fail with an UNKNOWN_ERROR error status.**

These kinds of error statuses are normally translated into HardwareError
exceptions by CoreTools.

Args:
  - uint16_t: The index of the node that you wish to inspect.

Returns:
  - char[20]: A binary node descriptor for the node in question.
"""

SG_INSPECT_STREAMER = RPCDeclaration(0x2017, "H", "L14s2x")
"""Get a binary streamer descriptor describing a streamer by its index.

This allows you to verify the action of the SG_ADD_STREAMER RPC by inspecting
the last streamer added to the sensor-graph.  The binary descriptor returned should
be identical to what you used to create it.

This RPC is also useful for inspecting what sensor-graph is programmed into a
given device by iterating over all streamers and returning their streamer descriptors.

Calling this RPC repeatedly along with SG_INSPECT_GRAPH_NODE will let you dump
a sensor-graph out of an IOTile device.

Args:
  - uint16_t: The index of the streamer that you wish to dump.

Returns:
  - uint32_t: An error code.
  - char[14]: A binary streamer descriptor suitable for passing to SG_ADD_STREAMER.

Possible Errors:
  - STREAMER_NOT_ALLOCATED: If the index corresponds to a streamer that has
    not been allocated.
"""



"""Base class for virtual devices designed to emulate physical devices."""

import logging
from iotile.core.exceptions import DataError
from iotile.core.hw.virtual import VirtualIOTileDevice
from iotile.core.hw.virtual.common_types import pack_rpc_payload, unpack_rpc_payload, AsynchronousRPCResponse
from .emulation_mixin import EmulationMixin
from .state_log import EmulationStateLog
from ..constants.rpcs import RPCDeclaration
from ..internal import EmulationLoop, AwaitableResponse
from ..utilities import format_rpc


#pylint:disable=abstract-method;This is an abstract base class
class EmulatedDevice(EmulationMixin, VirtualIOTileDevice):
    """Base class for virtual devices designed to emulate physical devices.

    This class adds additional state and test scenario loading functionality
    as well as tracing of state changes on the emulated device for comparison
    and verification purposes.

    Since all IOTile devices have a single TileBus interface that serializes
    RPCs between tiles so that only one runs at a time, this class spawns a
    single background rpc to work through the queue of RPCs that should be
    sent, sending them one at a time.  The rpc dispatch queue is started when
    start() is called and stopped synchronously when stop() is called.

    Args:
        iotile_id (int): A 32-bit integer that specifies the globally unique ID
            for this IOTile device.
        name (string): The 6 byte name that should be returned when anyone asks
            for the controller's name of this IOTile device using an RPC
    """

    __NO_EXTENSION__ = True

    def __init__(self, iotile_id, name):
        self.state_history = EmulationStateLog()

        VirtualIOTileDevice.__init__(self, iotile_id, name)
        EmulationMixin.__init__(self, None, self.state_history)

        self._logger = logging.getLogger(__name__)
        self.emulator = EmulationLoop(self._dispatch_rpc)

    def _dispatch_rpc(self, address, rpc_id, arg_payload):
        """Background work queue handler to dispatch RPCs."""

        try:
            # Send the RPC immediately and wait for the response
            resp = super(EmulatedDevice, self).call_rpc(address, rpc_id, arg_payload)
            self._track_change('device.rpc_sent', (address, rpc_id, arg_payload, resp, None), formatter=format_rpc)

            return resp
        except AsynchronousRPCResponse:
            self._track_change('device.rpc_started', (address, rpc_id, arg_payload, None, None), formatter=format_rpc)
            raise
        except Exception as exc:
            self._track_change('device.rpc_exception', (address, rpc_id, arg_payload, None, exc), formatter=format_rpc)
            raise

    def finish_async_rpc(self, address, rpc_id, response):
        """Finish a previous asynchronous RPC.

        This method should be called by a peripheral tile that previously
        had an RPC called on it and chose to response asynchronously by
        raising ``AsynchronousRPCResponse`` in the RPC handler itself.

        The response passed to this function will be returned to the caller
        as if the RPC had returned it immediately.

        The rpc response will be sent in the RPC thread.  By default this
        method will block until the response is finished.  If you don't
        want to block, you can pass sync=False

        Args:
            address (int): The tile address the RPC was called on.
            rpc_id (int): The ID of the RPC that was called.
            response (bytes): The bytes that should be returned to
                the caller of the RPC.
        """

        try:
            self.emulator.finish_async_rpc(address, rpc_id, response)
            self._track_change('device.rpc_finished', (address, rpc_id, None, response, None), formatter=format_rpc)
        except Exception as exc:
            self._track_change('device.rpc_exception', (address, rpc_id, None, response, exc), formatter=format_rpc)
            raise

    def start(self, channel=None):
        """Start this emulated device.

        This triggers the controller to call start on all peripheral tiles in
        the device to make sure they start after the controller does and then
        it waits on each one to make sure they have finished initializing
        before returning.

        Args:
            channel (IOTilePushChannel): the channel with a stream and trace
                routine for streaming and tracing data through a VirtualInterface
        """

        super(EmulatedDevice, self).start(channel)
        self.emulator.start()

    def stop(self):
        """Stop running this virtual device including any worker threads."""

        self.emulator.stop()

        super(EmulatedDevice, self).stop()

    def dump_state(self):
        """Dump the current state of this emulated object as a dictionary.

        Returns:
            dict: The current state of the object that could be passed to load_state.
        """

        state = {}

        state['tile_states'] = {}

        for address, tile in self._tiles.items():
            state['tile_states'][address] = tile.dump_state()

        return state

    def rpc(self, address, rpc_id, *args, **kwargs):
        """Immediately dispatch an RPC inside this EmulatedDevice.

        This function is meant to be used for testing purposes as well as by
        tiles inside a complex EmulatedDevice subclass that need to
        communicate with each other.  It should only be called from the main
        virtual device thread where start() was called from.

        **Background workers may not call this method since it may cause them to deadlock.**

        Args:
            address (int): The address of the tile that has the RPC.
            rpc_id (int): The 16-bit id of the rpc we want to call
            *args: Any required arguments for the RPC as python objects.
            **kwargs: Only two keyword arguments are supported:
                - arg_format: A format specifier for the argument list
                - result_format: A format specifier for the result

        Returns:
            list: A list of the decoded response members from the RPC.
        """

        if isinstance(rpc_id, RPCDeclaration):
            arg_format = rpc_id.arg_format
            resp_format = rpc_id.resp_format
            rpc_id = rpc_id.rpc_id
        else:
            arg_format = kwargs.get('arg_format', None)
            resp_format = kwargs.get('resp_format', None)

        arg_payload = b''

        if arg_format is not None:
            arg_payload = pack_rpc_payload(arg_format, args)

        self._logger.debug("Sending rpc to %d:%04X, payload=%s", address, rpc_id, args)

        resp_payload = self.call_rpc(address, rpc_id, arg_payload)
        if resp_format is None:
            return []

        resp = unpack_rpc_payload(resp_format, resp_payload)
        return resp

    def call_rpc(self, address, rpc_id, payload=b""):
        """Call an RPC by its address and ID.

        This will send the RPC to the background rpc dispatch thread and
        synchronously wait for the response.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes

        Returns:
            bytes: The response payload from the RPC
        """

        return self.emulator.call_rpc_external(address, rpc_id, payload)

    def trace_sync(self, data, timeout=5.0):
        """Send tracing data and wait for it to finish.

        This awaitable coroutine wraps VirtualIOTileDevice.trace() and turns
        the callback into an awaitable object.  The appropriate usage of this
        method is by calling it inside the event loop as:

        await device.trace_sync(data)

        Args:
            data (bytes): The raw data that should be traced.
            timeout (float): The maximum number of seconds to wait before
                timing out.

        Returns:
            awaitable: An awaitable object with the result.

            The result will be True if the data was sent successfully
            or False if the data could not be sent in its entirety.

            When False is returned, there is no guarantee about how much of
            the data was sent, if any, just that it was not known to be
            successfully sent.
        """

        done = AwaitableResponse()
        self.trace(data, callback=done.set_result)
        return done.wait(timeout)

    def stream_sync(self, report, timeout=120.0):
        """Send a report and wait for it to finish.

        This awaitable coroutine wraps VirtualIOTileDevice.stream() and turns
        the callback into an awaitable object.  The appropriate usage of this
        method is by calling it inside the event loop as:

        await device.stream_sync(data)

        Args:
            report (IOTileReport): The report that should be streamed.
            timeout (float): The maximum number of seconds to wait before
                timing out.

        Returns:
            awaitable: An awaitable object with the result.

            The result will be True if the data was sent successfully
            or False if the data could not be sent in its entirety.

            When False is returned, there is no guarantee about how much of
            the data was sent, if any, just that it was not known to be
            successfully sent.
        """

        done = AwaitableResponse()
        self.stream(report, callback=done.set_result)
        return done.wait(timeout)

    def synchronize_task(self, func, *args, **kwargs):
        """Run callable in the rpc thread and wait for it to finish.

        The callable ``func`` will be passed into the EmulationLoop and run
        there.  This method will block until ``func`` is finished and
        return/raise whatever that callable returns/raises.

        This method is mainly useful for performing an activity that needs to
        be synchronized with the rpc thread for safety reasons.

        If this method is called from the rpc thread itself, it will just
        run the task and return its result.

        Args:
            func (callable): A method with signature callable(*args, **kwargs),
                that will be called with the optional *args and **kwargs passed
                to this method.
            *args: Arguments that will be passed to callable.
            **kwargs: Keyword arguments that will be passed to callable.

        Returns:
            object: Whatever callable returns after it runs.
        """

        async def _runner():
            return func(*args, **kwargs)

        return self.emulator.run_task_external(_runner())

    def wait_idle(self):
        """Wait until the emulated device is idle.

        An idle device is one where there are no tasks pending on the
        emulation thread. This is different from wait_deferred_rpcs() a given
        rpc that was scheduled before your call to wait_deferred_rpcs() could
        have scheduled an rpc to run after your call.

        This method gives you a deterministic way to wait until there is no
        more activity on the device.  This is particularly useful for
        deterministic testing because you can start an action and know
        precisely when all potential chain reactions have stopped.

        This method will block the calling thread until the _rpc_queue is
        idle.

        **Calling wait_idle from the emulation thread would deadlock and will
        raise an exception.**
        """

        self.emulator.wait_idle()

    def restore_state(self, state):
        """Restore the current state of this emulated device.

        Args:
            state (dict): A previously dumped state produced by dump_state.
        """

        tile_states = state.get('tile_states', {})

        for address, tile_state in tile_states.items():
            address = int(address)
            tile = self._tiles.get(address)
            if tile is None:
                raise DataError("Invalid dumped state, tile does not exist at address %d" % address, address=address)

            tile.restore_state(tile_state)

    def load_metascenario(self, scenario_list):
        """Load one or more scenarios from a list.

        Each entry in scenario_list should be a dict containing at least a
        name key and an optional tile key and args key.  If tile is present
        and its value is not None, the scenario specified will be loaded into
        the given tile only.  Otherwise it will be loaded into the entire
        device.

        If the args key is specified is will be passed as keyword arguments
        to load_scenario.

        Args:
            scenario_list (list): A list of dicts for each scenario that should
                be loaded.
        """

        for scenario in scenario_list:
            name = scenario.get('name')
            if name is None:
                raise DataError("Scenario in scenario list is missing a name parameter", scenario=scenario)

            tile_address = scenario.get('tile')
            args = scenario.get('args', {})

            dest = self
            if tile_address is not None:
                dest = self._tiles.get(tile_address)

                if dest is None:
                    raise DataError("Attempted to load a scenario into a tile address that does not exist", address=tile_address, valid_addresses=list(self._tiles))

            dest.load_scenario(name, **args)



"""A reference device with properties for inspecting what has been set by update scripts."""

import base64
import logging
import asyncio
from monotonic import monotonic
from iotile.core.exceptions import ArgumentError, DataError
from ..virtual import EmulatedDevice, EmulatedPeripheralTile
from ..constants import rpcs, streams
from .reference_controller import ReferenceController


class ReferenceDevice(EmulatedDevice):
    """A reference implementation of an IOTile device.

    This is useful for testing the effects of updates and other build
    automation processes.

    Args:
        args (dict): A dictionary of optional creation arguments.  Currently
            supported are:
                iotile_id (int or hex string): The id of this device. This
                defaults to 1 if not specified.
    """

    __NO_EXTENSION__ = True

    STATE_NAME = "reference_device"
    STATE_VERSION = "0.1.0"

    def __init__(self, args):
        iotile_id = args.get('iotile_id', 1)
        controller_name = args.get('controller_name', 'refcn1')

        if isinstance(iotile_id, str):
            iotile_id = int(iotile_id, 16)

        super(ReferenceDevice, self).__init__(iotile_id, controller_name)

        self.controller = ReferenceController(8, {'name': controller_name}, device=self)
        self.add_tile(8, self.controller)
        self.reset_count = 0
        self._logger = logging.getLogger(__name__)
        self._simulating_time = args.get('simulate_time', True)
        self._accelerating_time = args.get('accelerate_time', False)

    async def _time_ticker(self):
        start = monotonic()
        counter = 0
        while self._simulating_time:
            self.controller.clock_manager.handle_tick()
            await self.emulator.wait_idle()

            counter += 1
            next_tick = start + counter
            if not self._accelerating_time:
                now = monotonic()

                while now < next_tick:
                    if not self._simulating_time:
                        return

                    step = min(next_tick - now, 0.1)
                    await asyncio.sleep(step)
                    now = monotonic()

        self._logger.debug("Time ticker task stopped due to _simulating_time flag cleared")

    def iter_tiles(self, include_controller=True):
        """Iterate over all tiles in this device in order.

        The ordering is by tile address which places the controller tile
        first in the list.

        Args:
            include_controller (bool): Include the controller tile in the
                results.

        Yields:
            int, EmulatedTile: A tuple with the tile address and tile object.
        """

        for address, tile in sorted(self._tiles.items()):
            if address == 8 and not include_controller:
                continue

            yield address, tile

    def start(self, channel=None):
        """Start this emulated device.

        This triggers the controller to call start on all peripheral tiles in
        the device to make sure they start after the controller does and then
        it waits on each one to make sure they have finished initializing
        before returning.

        Args:
            channel (IOTilePushChannel): the channel with a stream and trace
                routine for streaming and tracing data through a VirtualInterface
        """

        super(ReferenceDevice, self).start(channel)

        try:
            self.controller.start(channel)

            # Guarantee an initialization order so that our trace files are deterministic
            for address, tile in sorted(self._tiles.items()):
                if address == 8:
                    continue

                if not isinstance(tile, EmulatedPeripheralTile):
                    raise DataError("An emulated ReferenceDevice can only have a single controller and all other tiles must inherit from EmulatedPeripheralTile",
                                    address=address)

                tile.start(channel)

            async def _launch_tiles():
                await self.controller.reset()
                await asyncio.wait_for(self.controller.initialized.wait(), 2.0)

                # Note that we do not explicitly reset the tiles.
                # The controller resets all tiles in its reset method.
                for address, tile in sorted(self._tiles.items()):
                    if address == 8:
                        continue

                    await asyncio.wait_for(tile.initialized.wait(), 2.0)

            self.emulator.run_task_external(_launch_tiles())

            if self._simulating_time:
                self.emulator.add_task(None, self._time_ticker())
        except:
            self.stop()
            raise

    def stop(self):
        """Stop this emulated device."""

        self._simulating_time = False
        super(ReferenceDevice, self).stop()

    def open_streaming_interface(self):
        """Called when someone opens a streaming interface to the device.

        This method will automatically notify sensor_graph that there is a
        streaming interface opened.

        Returns:
            list: A list of IOTileReport objects that should be sent out
                the streaming interface.
        """

        super(ReferenceDevice, self).open_streaming_interface()

        self.rpc(8, rpcs.SG_GRAPH_INPUT, 8, streams.COMM_TILE_OPEN)
        return []

    def close_streaming_interface(self):
        """Called when someone closes the streaming interface to the device.

        This method will automatically notify sensor_graph that there is a no
        longer a streaming interface opened.
        """

        super(ReferenceDevice, self).close_streaming_interface()

        self.rpc(8, rpcs.SG_GRAPH_INPUT, 8, streams.COMM_TILE_CLOSED)

    def dump_state(self):
        """Dump the current state of this emulated object as a dictionary.

        Note that dump_state happens synchronously in the emulation thread to
        avoid any race conditions with accessing data members and ensure a
        consistent view of all state data.

        Returns:
            dict: The current state of the object that could be passed to load_state.
        """

        # Dump the state of all of the tiles
        def _background_dump():
            state = super(ReferenceDevice, self).dump_state()

            state['state_name'] = self.STATE_NAME
            state['state_version'] = self.STATE_VERSION
            state['reset_count'] = self.reset_count
            state['received_script'] = base64.b64encode(self.script).decode('utf-8')

            return state

        return self.synchronize_task(_background_dump)

    def restore_state(self, state):
        """Restore the current state of this emulated device.

        Note that restore_state happens synchronously in the emulation thread
        to avoid any race conditions with accessing data members and ensure a
        consistent atomic restoration process.

        This method will block while the background restore happens.

        Args:
            state (dict): A previously dumped state produced by dump_state.
        """

        state_name = state.get('state_name')
        state_version = state.get('state_version')

        if state_name != self.STATE_NAME or state_version != self.STATE_VERSION:
            raise ArgumentError("Invalid emulated device state name or version", found=(state_name, state_version),
                                expected=(self.STATE_NAME, self.STATE_VERSION))

        def _background_restore():
            # Restore the state of all of the tiles
            super(ReferenceDevice, self).restore_state(state)

            self.reset_count = state.get('reset_count', 0)
            self.script = base64.b64decode(state.get('received_script'))

        self.synchronize_task(_background_restore)



"""A generic RPC-level emulator for any IOTile device.

This package provides a framework that lets you emulate any IOTile device.
It includes:

- A generic emulation system based on the EmulationLoop class.  This is the
  basis for the emulation and allows running coroutines simulating tiles
  inside a single event loop and communication with those coroutines using
  RPC.
- A VirtualIOTileDevice subclass EmulatedDevice that includes the
  EmulationLoop and wraps it in an interface that can be handled to any
  VirtualInterface and served like a normal virtual device.  This means that
  you can make an emulated device available over any protocol that has a
  virtual interface such as bluetooth, websockets or MQTT.
- A reference implemention of the IOTile bus controller in Python, in the
  ReferenceController class.  This allows for the emulation of physical IOTile
  devices by just writing small classes that emulate the behavior of the
  peripheral tiles.  The ReferenceDevice class is the correct base class for
  these endeavors.
- A device adapter that allows you to directly run an EmulatedDevice inside
  of the iotile tool for testing and demos.
"""

from .virtual import EmulatedDevice, EmulatedTile, EmulatedPeripheralTile
from . import constants

__all__ = ['EmulatedTile', 'EmulatedDevice', 'EmulatedPeripheralTile', 'constants']



"""RPCs implemented by the TileManager subsystem on an IOTile controller.

These RPCs allow querying information on any registered tiles (including the
controller itself which lists itself first).  The TileManager subsystem is
also involved in bootloading new firmware onto peripheral tiles so there are
RPCs for tiles that are instructed to bootload as a result of their call to
REGISTER_TILE to use to notify the controller of their bootloading status and
request new chunks of firmware.
"""

from .rpc_declaration import RPCDeclaration

REGISTER_TILE = RPCDeclaration(0x2a00, "3B6s6BBL", "HHH")
"""Register a peripheral tile with the local controller.

This RPC is called by all peripheral tiles when they first boot up in order to
check in with the IOTile controller on their local bus.  The tiles send
information about themselves that the controller caches in its tile_manager
subsystem.

Args:
  - uint8_t: a numerical hardware type identifer for the processor the tile is
    running on.
  - two bytes: (major, minor) API level of the tile executive running on this
    tile.  The api version is encoded with the major version as a byte
    followed by the minor version.
  - 6-character string: The tile firmware identifier that is used to match
    it with the correct proxy module by HardwareManager.
  - 3 bytes: (major, minor, patch) version identifier of the application
    firmware running on this tile.
  - 3 bytes: (major, minor, patch) version identifier of the tile executive
    running on this tile.
  - uint8_t: The slot number that refers to the physical location of this
    tile on the board.  It is used to assign the tile a fixed address on
    the TileBus.
  - uint32_t: A unique identifier for this tile, if the tile supports
    unique identifiers.

Returns:
  - uint16_t: The address that the tile was assigned based on the slot informatioN
    that it passed to the controller.
  - uint16_t: The run-level for the tile.  The tile executive uses this information
    to determine whether it should pass control to application firmware or prevent
    any user application code from running because the device is in safe mode.
  - uint16_t: Whether the tile should run in debug mode.  This hint tells the
    tile firmware that it should ensure that it runs in a debug-compatible
    mode, which may among other things entail using lighter sleep states that
    don't break attached debuggers.
"""

COUNT_TILES = RPCDeclaration(0x2a01, "", "H")
"""Count how many tiles have registered themselves with the controller.

This RPC is implemented by IOTile controllers and forms part of the
TileManager subsystem.  It is used to allow iteration over the TileManager's
cache of tile registration data by listing how many tiles are in the cache.

The count of registered tiles can never decrease unless the controller resets
since there is no UNREGISTER_TILE command.

Returns:
 - uint16_t: The number of currently registered tiles.
"""

DESCRIBE_TILE = RPCDeclaration(0x2a02, "H", "3B6s6BBL")
"""Describes the tile at the given index in the TileManager cache.

The information returned is the exact same data as what REGISTER_TILE
includes as its argument.

Args:
  - uint16_t: The index of the tile you wish to describe.  Must be less than
    what COUNT_TILES returns.

Returns:
  - uint8_t: a numerical hardware type identifer for the processor the tile is
    running on.
  - two bytes: (major, minor) API level of the tile executive running on this
    tile.  The api version is encoded with the major version as a byte
    followed by the minor version.
  - 6-character string: The tile firmware identifier that is used to match it
    with the correct proxy module by HardwareManager.
  - 3 bytes: (major, minor, patch) version identifier of the application
    firmware running on this tile.
  - 3 bytes: (major, minor, patch) version identifier of the tile executive
    running on this tile.
  - uint8_t: The slot number that refers to the physical location of this tile
    on the board.  It is used to assign the tile a fixed address on the
    TileBus.
  - uint32_t: A unique identifier for this tile, if the tile supports unique
    identifiers.
"""

GET_TILE_STATE = RPCDeclaration(0x2a03, "H", "BB")
"""Get the current address and state of a tile by index.

This RPC queries the TileManager cache for the assigned address and
state of the tile at the given index.

Args:
  - uint16_t: The index of the tile you wish to describe.  Must be less than
    what COUNT_TILES returns.

Returns:
  - uint8_t: The address that the tile was assigned.
  - uint8_t: The tile's state, which is taken from the TILE_STATE enum to
    indicate if it is running, dormant, registered, bootloading, etc.
"""

BOOTLOAD_NOTIFY_STARTED = RPCDeclaration(0x2a04, "", "")
"""Notify the controller that the sender has started bootloading.

Only a single tile is allowed to bootload at a time so no parameters are sent
indicating which tile has started bootloading.  TileManager maintains internal
state indicating which tile, if any, is supposed to bootload.

If no tile is supposed to bootload, the RPC fails with the UNKNOWN error code.
"""

BOOTLOAD_POLL_FIRMWARE = RPCDeclaration(0x2a05, "", "")
"""Request a chunk of firmware starting at a specific address from the controller.

Tiles that are in bootloading mode must repeatedly call this function to read in
their new firmware image chunk by chunk, incrementing address with each call.

It is not required that the tile's read their firmware chunks in order.

Args:
  - uint32_t: The address of the firmware chunk that you wish to receive.

Returns:
  - char[]: Up to 20 bytes of firmware starting at the given address.  May be
    less if the firmware image ends less than 20 bytes after the given
    address.
"""

BOOTLOAD_NOTIFY_ENDED = RPCDeclaration(0x2a06, "", "")
"""Notify the controller that the sender has finished bootloading.

Only a single tile is allowed to bootload at a time so no parameters are sent
indicating which tile has started bootloading.  TileManager maintains internal
state indicating which tile, if any, is supposed to bootload.

If no tile is supposed to bootload, the RPC fails with the UNKNOWN error code.
"""



""" A simple stream manager subsystem.

This subsystem is appropriate primarily for use with the emulated device adapter.
It supports receiving reports from the sensor-graph subsystem and queuing them
for transmission out of the streaming interface.

TODO:
  - [ ] Keep track of last attempt and last success times for streamers
"""


import logging
from iotile.core.exceptions import InternalError
from iotile.core.hw.reports import IndividualReadingReport
from .controller_system import ControllerSubsystemBase


class QueuedStreamer(object):
    def __init__(self, streamer, callback):
        self.status = 1
        self.streamer = streamer
        self.callback = callback
        self.initial_count = streamer.walker.count()

        self.sent_count = 0
        self.highest_ack = 0

    @property
    def remaining(self):
        return self.initial_count - self.sent_count

    def record_acknowledgement(self, ack):
        if ack > self.highest_ack:
            self.highest_ack = ack


class BasicStreamingSubsystem(ControllerSubsystemBase):
    """Container for a very basic streaming subsystem."""

    MAX_REPORT_SIZE = 3*64*1024

    def __init__(self, emulator, device, id_assigner):
        super(BasicStreamingSubsystem, self).__init__(emulator)

        self._in_progress_streamers = set()
        self._queue = emulator.create_queue(register=False)
        self._actively_streaming = False
        self._device = device
        self._logger = logging.getLogger(__name__)
        self._streamer_queue = []
        self._id_assigner = id_assigner

        self.get_timestamp = lambda: 0
        self.get_uptime = lambda: 0

    async def _reset_vector(self):
        #FIXME: Clear all queued streamers on a reset

        self.initialized.set()

        while True:
            queued = await self._queue.get()

            try:
                success = False
                while queued.remaining > 0:
                    report, num_readings, highest_id = self._build_report(queued.streamer)
                    queued.sent_count += num_readings

                    if isinstance(report, IndividualReadingReport):
                        queued.record_acknowledgement(highest_id)

                    self._logger.debug("Streamer %d: starting with report %s", queued.streamer.index, report)
                    success = await self._device.stream_sync(report)
                    self._logger.debug("Streamer %d: finished with success=%s", queued.streamer.index, success)
            except:  #pylint:disable=bare-except;This background worker should never die
                self._logger.exception("Exception during streaming of streamer %s", queued.streamer)
                success = False
            finally:
                self._in_progress_streamers.remove(queued.streamer.index)
                self._logger.debug("Current in_progress is %s", self._in_progress_streamers)

                if queued.callback is not None:
                    try:
                        queued.callback(queued.streamer.index, success, queued.highest_ack)
                    except: #pylint:disable=bare-except;This background worker should never die
                        self._logger.exception("Exception during completion callback of streamer %s", queued.streamer)

    def in_progress(self):
        """Get a set of in progress streamers."""

        return set(self._in_progress_streamers)

    def clear_to_reset(self, config_vars):
        """Clear all volatile information across a reset."""

        super(BasicStreamingSubsystem, self).clear_to_reset(config_vars)
        self._in_progress_streamers = set()

    def _build_report(self, streamer):
        report_id = None
        if streamer.requires_id():
            report_id = self._id_assigner()

        return streamer.build_report(self._device.iotile_id, self.MAX_REPORT_SIZE, device_uptime=self.get_uptime(), report_id=report_id)

    def process_streamer(self, streamer, callback=None):
        """Start streaming a streamer.

        Args:
            streamer (DataStreamer): The streamer itself.
            callback (callable): An optional callable that will be called as:
                callable(index, success, highest_id_received_from_other_side)
        """

        index = streamer.index

        if index in self._in_progress_streamers:
            raise InternalError("You cannot add a streamer again until it has finished streaming.")

        queue_item = QueuedStreamer(streamer, callback)
        self._in_progress_streamers.add(index)

        self._logger.debug("Streamer %d: queued to send %d readings", index, queue_item.initial_count)
        self._queue.put_nowait(queue_item)


class StreamingSubsystemMixin(object):
    """Mixin for an IOTileController that implements the streaming subsystem.

    Depending on the desired level of realism, you can choose to use either a
    complete streaming emulation that will send each streaming report chunk by
    chunk with RPCs to a destination tile, or you can opt for the simpler
    basic streaming subsystem that just takes the report and queues it for
    processing.

    Args:
        basic (bool): Whether to use a basic model of report streaming or the
            complete simulation.  Defaults to True, which means basic.
    """

    def __init__(self, emulator, basic=True):
        if not basic:
            raise InternalError("Full stream manager support is not yet implemented")

        self.stream_manager = BasicStreamingSubsystem(emulator, self._device, self.sensor_log.allocate_id)
        self._post_config_subsystems.append(self.stream_manager)



"""These RPCs implement the configuration variable interface of any tile.

Config variables are used to stream configuration information to a tile when
it resets, before any application firmware is run.  All tiles must implement
these RPCs in order to allow for boot-time configuration.

The local IOTile controller is in charge of persisting and streaming the
actual config variables to each tile, these RPCs just provide the API that it
can use in order to do so.
"""

from .rpc_declaration import RPCDeclaration


LIST_CONFIG_VARIABLES = RPCDeclaration(10, "H", "H9H")
"""Iteratively list all defined config variables on this tile.

This method returns a chunk of the tile's internal config variable declaration
table.  You pass an offset as an argument and it returns the config_variable
ids of the next 9 config variables prepended with a valid count.  The return
value of this RPC is fixed size so the count value will tell you how many
entries are valid.

Clients should call this function repeatedly incrementing the offset until
count is returned as 0 indicating that the table has been exhausted.

Args:
  - uint16_t: An offset into the tile's config variable table.

Returns:
  - uint16_t: The count of valid config variable ids that follow.  This will
    always be [0, 9] inclusive.
  - uint16_t[9]: An array of 16 bit config variable ids.  Only the first count
    entries are valid and the order is unspecified but usually sorted lowest
    to highest.
"""

DESCRIBE_CONFIG_VARIABLE = RPCDeclaration(11, "H", "HHLHH")
"""Describe the type, size and internal address of a config variable by id.

This method returns the internal config variable descriptor for a given
config variable by its 16-bit config id.  When an error is returned the
contents of the rest of the response are undefined.

Args:
  - uint16_t: The config variable ID you wish to inspect.  This should
    be in the list returned by LIST_CONFIG_VARIABLES or an error will
    be returned.

Returns:
  - uint16_t: An error code, otherwise NO_ERROR to indicate success.
  - uint16_t: Reserved for alignment, value is undefined.
  - uint32_t: The internal RAM address of this config variable.  There
    is typically no need to use this value for anything but it could
    be useful for advanced debugging scenarios.
  - uint16_t: The config variable id of this variable, will always
    match what you passed in the arguments.
  - uint16_t: Encoded bit field.  The low 15 bits are the total
    size of the space allocated for this config variable in bytes.
    The high bit indicates whether this is an array or a single
    value.

Possible Errors:
  - INVALID_ARRAY_KEY: The config variable id could not be found
"""

SET_CONFIG_VARIABLE = RPCDeclaration(12, "HHV", "H")
"""Set a chunk (possibly all) of the config variable's value.

Config variables for the purposes of set and get are treated as binary strings
of bytes.  This function lets you stream in up to 20 bytes of data into the
config variable at the specified offset.

Variable length config variables keep track of the highest offset written to
by a call SET_CONFIG_VARIABLE and report that as their size.  For that reason
it is important to send data to a long config variable from low to high
offsets so that the final length is correct.  Writing to a low offset after
having previously written to a high offset will truncate the stored data.

If an error is returned, the state of the config variable is unchanged by this
call.

Args:
  - uint16_t: the id of the config variable that you want to set.
  - uint16_t: The offset that you want to start writing at.
  - char[]: Variable length buffer up to the end of the RPC payload that will
    be copied into the variable's value starting at the specified offset.

Returns:
  - uint16_t: An error code.

Possible Errors:
  - INVALID_ARRAY_KEY: The config variable id could not be found
  - STATE_CHANGE_AT_INVALID_TIME: This RPC was called after a peripheral
    tile's firmware had already started execution. Config variable are not
    allowed to change while a tile is running.
  - INPUT_BUFFER_TOO_LONG: The data would overflow the space allocated for the
    config variable.
"""

GET_CONFIG_VARIABLE = RPCDeclaration(13, "HH", "V")
"""Get a chunk (possibly all) of the config variable's value.

Config variables for the purposes of set and get are treated as binary strings
of bytes.  This function lets you get up to 20 bytes of data into the config
variable at the specified offset.

Callers should call this function repeatedly with incrementing offset until
it returns 0 bytes indicating that you have received all data from that
config variable.

There is no way for the tile to communicate to you that the call failed except
by returning a 0-length result so you cannot tell the difference with this
call between a config variable that does not exist and one that has variable
size and no data written yet.

Fixed size config variables will always return their fixed size regardless of
how much data has been written to them.

Args:
  - uint16_t: the id of the config variable that you want to set.
  - uint16_t: The offset that you want to start writing at.

Returns:
  - char[]: Up to 20 bytes of data from the config variable starting at the
    given offset.  The return value may be smaller than 20 bytes if the end
    of the data is reached.
"""



"""Helper class for automatically dumping and restoring all member variables."""

import inspect
from iotile.core.exceptions import ArgumentError, DataError


class SerializableState(object):
    """A helper class that can dump itself to a dictionary.

    This class only dumps and restores public data members that it discovers
    using the inspect module.  If the datatype of the public member is such
    that it needs preprocessing before being json serializable, you can pass
    its name to mark_complex(name, serializer, deserializer) and those functions
    will be called on dump and restore to serialize/deserialize the property.
    """

    def __init__(self):
        self._complex_properties = {}
        self._ignored_properties = set()

    def dump(self):
        """Serialize this state by iterating over all public properties using inspect.

        Returns:
            dict: The serialized representation of this state.
        """

        return {x: self.dump_property(x) for x in self.get_properties()}

    def restore(self, state):
        """Restore this state from the output of a previous call to dump().

        Only those properties in this object and listed in state will be
        updated.  Other properties will not be modified and state may contain
        keys that do not correspond with properties in this object.

        Args:
            state (dict): A serialized representation of this object.
        """

        own_properties = set(self.get_properties())
        state_properties = set(state)

        to_restore = own_properties.intersection(state_properties)

        for name in to_restore:
            value = state.get(name)

            if name in self._complex_properties:
                value = self._complex_properties[name][1](value)

            setattr(self, name, value)

    def mark_ignored(self, name):
        """Make a property that should not be serialized/deserialized.

        Args:
            name (str): The name of the property to ignore.
        """

        self._ignored_properties.add(name)

    def mark_complex(self, name, serializer, deserializer):
        """Mark a property as complex with serializer and deserializer functions.

        Args:
            name (str): The name of the complex property.
            serializer (callable): The function to call to serialize the property's
                value to something that can be saved in a json.
            deserializer (callable): The function to call to unserialize the property
                from a dict loaded by a json back to the original value.
        """

        self._complex_properties[name] = (serializer, deserializer)

    def mark_typed_list(self, name, type_object):
        """Mark a property as containing serializable objects of a given type.

        This convenience method allows you to avoid having to call
        ``mark_complex()`` whenever you need to serialize a list of objects.
        This method requires that all members of the given list be of a single
        class that contains a dump() method and a Restore() class method where
        type_object.Restore(x.dump()) == x.

        Args:
            name (str): The name of the complex property.
            type_object: The class object that will be contained inside
                this list.
        """

        if not hasattr(type_object, 'dump'):
            raise ArgumentError("The passed type object %s is missing required method: dump()" % type_object)
        if not hasattr(type_object, 'Restore'):
            raise ArgumentError("The passed type object %s is missing required method: Restore()" % type_object)

        def _dump_list(obj):
            if obj is None:
                return None

            if not isinstance(obj, list):
                raise DataError("Property %s marked as list was not a list: %s" % (name, repr(obj)))

            return [x.dump() for x in obj]

        def _restore_list(obj):
            if obj is None:
                return obj

            return [type_object.Restore(x) for x in obj]

        self.mark_complex(name, _dump_list, _restore_list)

    def mark_typed_map(self, name, type_object):
        """Mark a property as containing a map str to serializable object.

        This convenience method allows you to avoid having to call
        ``mark_complex()`` whenever you need to serialize a dict of objects.
        This method requires that all members of the given dict be of a single
        class that contains a dump() method and a Restore() class method where
        type_object.Restore(x.dump()) == x.

        Args:
            name (str): The name of the complex property.
            type_object: The class object that will be contained inside
                this dict.
        """

        if not hasattr(type_object, 'dump'):
            raise ArgumentError("The passed type object %s is missing required method: dump()" % type_object)
        if not hasattr(type_object, 'Restore'):
            raise ArgumentError("The passed type object %s is missing required method: Restore()" % type_object)

        def _dump_map(obj):
            if obj is None:
                return None

            if not isinstance(obj, dict):
                raise DataError("Property %s marked as list was not a dict: %s" % (name, repr(obj)))

            return {key: val.dump() for key, val in obj.items()}

        def _restore_map(obj):
            if obj is None:
                return obj

            return {key: type_object.Restore(val) for key, val in obj.items()}

        self.mark_complex(name, _dump_map, _restore_map)

    def mark_typed_object(self, name, type_object):
        """Mark a property as containing a serializable object.

        This convenience method allows you to avoid having to call
        ``mark_complex()`` whenever you need to serialize a complex object.
        This method requires that property ``name`` be a single class that
        contains a dump() method and a Restore() class method where
        type_object.Restore(x.dump()) == x.

        Args:
            name (str): The name of the complex property.
            type_object: The class object that will be contained inside
                this property.
        """

        if not hasattr(type_object, 'dump'):
            raise ArgumentError("The passed type object %s is missing required method: dump()" % type_object)
        if not hasattr(type_object, 'Restore'):
            raise ArgumentError("The passed type object %s is missing required method: Restore()" % type_object)

        def _dump_obj(obj):
            if obj is None:
                return None

            return obj.dump()

        def _restore_obj(obj):
            if obj is None:
                return obj

            return type_object.Restore(obj)

        self.mark_complex(name, _dump_obj, _restore_obj)

    def dump_property(self, name):
        """Serialize a property of this class by name.

        Args:
            name (str): The name of the property to dump.

        Returns:
            object: The serialized value of the property.
        """

        if not hasattr(self, name):
            raise ArgumentError("Unknown property %s" % name)

        value = getattr(self, name)
        if name in self._complex_properties:
            value = self._complex_properties[name][0](value)

        return value

    def get_properties(self):
        """Get a list of all of the public data properties of this class.

        Returns:
            list of str: A list of all of the public properties in this class.
        """

        names = inspect.getmembers(self, predicate=lambda x: not inspect.ismethod(x))
        return [x[0] for x in names if not x[0].startswith("_") and x[0] not in self._ignored_properties]



"""RPCs that are implemented by a tilebus controller outside of a subsystem.

These are generic RPCs that set configuration information for the entire
device or store persistent metadata like app or os tags.
"""

from .rpc_declaration import RPCDeclaration


SET_OS_APP_TAG = RPCDeclaration(0x100B, "LLBB", "L")
"""Update the OS and/or APP tag of the device.

The OS and APP tags are 20 bit numerical identifiers that identify the
firmware and sensorgraph, respectively running on an IOTile device.  They have
an associated X.Y version number for each tag that identifies the version
of os or app running on the device.

Each tag is packed into a 32-bit (little endian) number in the following
way:

lowest 20 bits: the numerical tag identifier
6 bits: the minor version number
6 bits: the major version number

So the highest 12 bits are a 6.6 fixed point integer storing MAJOR.MINOR
version numbers and the lowest 20 bits are the numerical tag value itself.

This RPC can update both tags/versions at the same time, only one or neither
based on the flags included in the RPC call.

Args:
  - uint32_t: The new packed os tag and version.
  - uint32_t: The new packes app tag and version.
  - uint8_t: Update the OS tag, if 0 then the new OS tag is ignored and the
    current one stored is used.
  - uint8_t: Update the APP tag, if 0 then the new APP tag is ingored and
    the current one stored is used.

Returns:
  - uint32_t: An error code.  No error codes are currently possible.
"""



"""Tests of various utilities used in IOTileDevice emulation."""

from __future__ import print_function
import pytest
from iotile.core.exceptions import DataError, ArgumentError
from iotile.emulate.virtual.emulated_tile import parse_size_name, ConfigDescriptor


@pytest.mark.parametrize("type_name, return_tuple", [
    ("uint8_t", (1, 1, 'B', False)),
    ("uint8_t[15]", (15, 1, 'B', True)),
    ("int8_t", (1, 1, 'b', False)),
    ("int8_t[15]", (15, 1, 'b', True)),
    ("char", (1, 1, 'B', False)),
    ("char[15]", (15, 1, 'B', True)),
    ("uint16_t", (2, 2, 'H', False)),
    ("uint16_t[15]", (30, 2, 'H', True)),
    ("int16_t", (2, 2, 'h', False)),
    ("int16_t[15]", (30, 2, 'h', True)),
    ("uint32_t", (4, 4, 'L', False)),
    ("uint32_t[15]", (60, 4, 'L', True)),
    ("int32_t", (4, 4, 'l', False)),
    ("int32_t[15]", (60, 4, 'l', True))
])
def test_config_parsing(type_name, return_tuple):
    """Make sure we can properly parse all config variable types."""

    actual_tuple = parse_size_name(type_name)

    print(type_name)
    assert return_tuple == actual_tuple


@pytest.mark.parametrize("type_name, data, python_type, latched_value", [
    ('uint8_t', bytearray([1]), None, 1),
    ('uint8_t', bytearray([255]), None, 255),
    ('int8_t', bytearray([1]), None, 1),
    ('int8_t', bytearray([255]), None, -1),
    ('uint16_t', bytearray([1, 0]), None, 1),
    ('uint16_t', bytearray([255, 255]), None, 0xFFFF),
    ('uint16_t', bytearray([255]), None, 0xFF),  # Make sure size extension happens correctly
    ('int16_t', bytearray([1, 0]), None, 1),
    ('int16_t', bytearray([255, 255]), None, -1),
    ('uint32_t', bytearray([1, 0, 0, 0]), None, 1),
    ('uint32_t', bytearray([255, 255, 255, 255]), None, 0xFFFFFFFF),
    ('int32_t', bytearray([1, 0, 0, 0]), None, 1),
    ('int32_t', bytearray([255, 255, 255, 255]), None, -1),

    # Arrays
    ('uint16_t[5]', bytearray([1, 0, 2, 0, 3, 0, 4, 0]), None, [1, 2, 3, 4]),
    ('uint16_t[5]', bytearray([1, 0, 2, 0, 3, 0, 4]), None, [1, 2, 3, 4]),  # Make sure size extension happens correctly
    ('char[16]', bytearray(b'test string') + bytearray(1), "string", u"test string")
])
def test_config_latching(type_name, data, python_type, latched_value):
    """Make sure we can properly decode variables."""

    print("Testing config latching of type %s with data %s" % (type_name, repr(data)))
    desc = ConfigDescriptor(0x8000, type_name, python_type=python_type)
    desc.update_value(0, data)

    actual_latched = desc.latch()
    assert actual_latched == latched_value


def test_nonnull_string():
    """Make sure we throw an exception for non-null terminated strings."""
    desc = ConfigDescriptor(0x8000, 'char[16]', python_type='string')
    desc.update_value(0, b'test string')

    with pytest.raises(DataError):
        desc.latch()


def test_empty():
    """Make sure we throw en error if we latch an empty config var."""

    desc = ConfigDescriptor(0x8000, 'uint8_t')
    with pytest.raises(DataError):
        desc.latch()


def test_unsupported_type():
    """Make sure we throw en error if we have an unknown python type."""

    with pytest.raises(ArgumentError):
        ConfigDescriptor(0x8000, 'uint8_t', default=b'\0', python_type="unsupported")


@pytest.mark.parametrize("type_name, default_value, latched_value, python_type, expected_exc", [
    ("uint8_t", 15, 15, None, None),
    ("uint8_t[15]", (15, 20, 25), [15, 20, 25], None, None),
    ("uint8_t[15]", bytearray([15, 20, 25]), [15, 20, 25], None, None),
    ("char[16]", u'test', 'test', "string", None),
    ("char[16]", b'test', 'test', "string", None),
    ("char[16]", bytearray(b'test'), 'test', "string", DataError)
])
def test_default_values(type_name, default_value, latched_value, python_type, expected_exc):
    """Make sure we can properly convert default values to binary."""

    desc = ConfigDescriptor(0x8000, type_name, default_value, python_type=python_type)
    desc.clear()

    if expected_exc is not None:
        with pytest.raises(expected_exc):
            desc.latch()
    else:
        assert desc.latch() == latched_value


def test_bool_conversion():
    """Make sure we can convert to bool."""

    desc = ConfigDescriptor(0x8000, 'uint8_t', 0, python_type='bool')
    desc.clear()

    assert desc.latch() is False

    desc = ConfigDescriptor(0x8000, 'uint8_t', 1, python_type='bool')
    desc.clear()

    assert desc.latch() is True

    with pytest.raises(ArgumentError):
        desc = ConfigDescriptor(0x8000, 'uint8_t[2]', 0, python_type='bool')



"""A reference implementation of the iotile controller that allows update verification."""

import logging
import asyncio
from iotile.core.hw.virtual import tile_rpc, TileNotFoundError
from iotile.core.hw.reports import IOTileReading
from iotile.core.exceptions import ArgumentError
from iotile.sg.model import DeviceModel
from iotile.sg.parser import SensorGraphFileParser
from iotile.sg.optimizer import SensorGraphOptimizer
from ..virtual import EmulatedTile
from ..constants import rpcs, Error

from .controller_features import (RawSensorLogMixin, RemoteBridgeMixin,
                                  TileManagerMixin, ConfigDatabaseMixin, SensorGraphMixin,
                                  StreamingSubsystemMixin, ClockManagerMixin)


class ReferenceController(RawSensorLogMixin, RemoteBridgeMixin,
                          TileManagerMixin, ConfigDatabaseMixin,
                          SensorGraphMixin, StreamingSubsystemMixin,
                          ClockManagerMixin, EmulatedTile):
    """A reference iotile controller implementation.

    This tile implements the major behavior of an IOTile controller including
    setting and inspecting sensorgraphs as well as setting and inspecting user
    keys and app/os tags.  The default reference implementation does not do
    anything with set data except expose it on public properties so that we
    can verify what was done to the device.

    Args:
        address (int): The address of this tile in the VirtualIOTIleDevice
            that contains it
        args (dict): The arguments to use to create this tile.  The currently supported
            arguments are:
                name (str): The 6 character name that should be returned when this
                    tile is asked for its status to allow matching it with a proxy
                    object.
        device (TileBasedVirtualDevice) : optional, device on which this tile is running
    """

    STATE_NAME = "reference_controller"
    STATE_VERSION = "0.1.0"

    def __init__(self, address, args, device=None):
        self.name = args.get('name', 'refcn1')
        self._device = device
        self._logger = logging.getLogger(__name__)
        self._post_config_subsystems = []

        if not isinstance(self.name, bytes):
            self.name = self.name.encode('utf-8')

        model = DeviceModel()

        EmulatedTile.__init__(self, address, device)

        # Initialize all of the controller subsystems
        ClockManagerMixin.__init__(self, device.emulator, has_rtc=False) #FIXME: Load the controller model info to get whether it has an RTC
        ConfigDatabaseMixin.__init__(self, 4096, 4096)  #FIXME: Load the controller model info to get its memory map
        TileManagerMixin.__init__(self, device.emulator)
        RemoteBridgeMixin.__init__(self, device.emulator)
        RawSensorLogMixin.__init__(self, device.emulator, model)
        StreamingSubsystemMixin.__init__(self, device.emulator, basic=True)
        SensorGraphMixin.__init__(self, device.emulator, self.sensor_log, self.stream_manager, model=model)

        # Establish required post-init linkages between subsystems
        self.clock_manager.graph_input = self.sensor_graph.process_input
        self.sensor_graph.get_timestamp = self.clock_manager.get_time
        self.stream_manager.get_timestamp = self.clock_manager.get_time
        self.stream_manager.get_uptime = lambda: self.clock_manager.get_time(False)

        self.app_info = (0, "0.0")
        self.os_info = (0, "0.0")

        self.register_scenario('load_sgf', self.load_sgf)

    def _handle_reset(self):
        """Reset this controller tile.

        This process will call _handle_reset() for all of the controller
        subsystem mixins in order to make sure they all return to their proper
        reset state.

        It will then reset all of the peripheral tiles to emulate the behavior
        of a physical POD where the controller tile cuts power to all
        peripheral tiles on reset for a clean boot.

        This will clear all subsystems of this controller to their reset
        states.

        The order of these calls is important to guarantee that everything is
        in the correct state before resetting the next subsystem.

        The behavior of this function is different depending on whether
        deferred is True or False.  If it's true, this function will only
        clear the config database and then queue all of the config streaming
        rpcs to itself to load in all of our config variables.  Once these
        have been sent, it will reset the rest of the controller subsystems.
        """

        self._logger.info("Resetting controller")
        self._device.reset_count += 1

        super(ReferenceController, self)._handle_reset()

        # Load in all default values into our config variables before streaming
        # updated data into them.
        self.reset_config_variables()

    async def _reset_vector(self):
        """Initialize the controller's subsystems inside the emulation thread."""

        # Send ourselves all of our config variable assignments
        config_rpcs = self.config_database.stream_matching(8, self.name)
        for rpc in config_rpcs:
            await self._device.emulator.await_rpc(*rpc)

        config_assignments = self.latch_config_variables()
        self._logger.info("Latched config variables at reset for controller: %s", config_assignments)

        for system in self._post_config_subsystems:
            try:
                system.clear_to_reset(config_assignments)
                await asyncio.wait_for(system.initialize(), timeout=2.0)
            except:
                self._logger.exception("Error initializing %s", system)
                raise

        self._logger.info("Finished clearing controller to reset condition")

        # Now reset all of the tiles
        for address, _ in self._device.iter_tiles(include_controller=False):
            self._logger.info("Sending reset signal to tile at address %d", address)

            try:
                await self._device.emulator.await_rpc(address, rpcs.RESET)
            except TileNotFoundError:
                pass
            except:
                self._logger.exception("Error sending reset signal to tile at address %d", address)
                raise

        self.initialized.set()

    def dump_state(self):
        """Dump the current state of this emulated object as a dictionary.

        Returns:
            dict: The current state of the object that could be passed to load_state.
        """

        superstate = super(ReferenceController, self).dump_state()

        superstate.update({
            'state_name': self.STATE_NAME,
            'state_version': self.STATE_VERSION,
            'app_info': self.app_info,
            'os_info': self.os_info,

            # Dump all of the subsystems
            'remote_bridge': self.remote_bridge.dump(),
            'tile_manager': self.tile_manager.dump(),
            'config_database': self.config_database.dump(),
            'sensor_log': self.sensor_log.dump()
        })

        return superstate

    def restore_state(self, state):
        """Restore the current state of this emulated object.

        Args:
            state (dict): A previously dumped state produced by dump_state.
        """

        super(ReferenceController, self).restore_state(state)

        state_name = state.get('state_name')
        state_version = state.get('state_version')

        if state_name != self.STATE_NAME or state_version != self.STATE_VERSION:
            raise ArgumentError("Invalid emulated device state name or version", found=(state_name, state_version),
                                expected=(self.STATE_NAME, self.STATE_VERSION))

        self.app_info = state.get('app_info', (0, "0.0"))
        self.os_info = state.get('os_info', (0, "0.0"))

        # Notify all subsystems of our intent to restore in case they need to prepare
        self.sensor_log.prepare_for_restore()

        # Restore all of the subsystems
        self.remote_bridge.restore(state.get('remote_bridge', {}))
        self.tile_manager.restore(state.get('tile_manager', {}))
        self.config_database.restore(state.get('config_database', {}))
        self.sensor_log.restore(state.get('sensor_log', {}))

    @tile_rpc(*rpcs.HARDWARE_VERSION)
    def hardware_version(self):
        """Get a hardware identification string."""

        hardware_string = self.hardware_string

        if not isinstance(hardware_string, bytes):
            hardware_string = self.hardware_string.encode('utf-8')

        if len(hardware_string) > 10:
            self._logger.warn("Truncating hardware string that was longer than 10 bytes: %s", self.hardware_string)

        if len(hardware_string) < 10:
            hardware_string += b'\0'*(10 - len(hardware_string))

        return [hardware_string]

    @tile_rpc(0x1008, "", "L8xLL")
    def controller_info(self):
        """Get the controller UUID, app tag and os tag."""

        return [self._device.iotile_id, _pack_version(*self.os_info), _pack_version(*self.app_info)]

    @tile_rpc(*rpcs.SET_OS_APP_TAG)
    def set_app_os_tag(self, os_tag, app_tag, update_os, update_app):
        """Update the app and/or os tags."""

        update_os = bool(update_os)
        update_app = bool(update_app)

        if update_os:
            self.os_info = _unpack_version(os_tag)

        if update_app:
            self.app_info = _unpack_version(app_tag)

        return [Error.NO_ERROR]

    # FIXME: Move this to the background thread
    def load_sgf(self, sgf_data):
        """Load, persist a sensor_graph file.

        The data passed in `sgf_data` can either be a path or the already
        loaded sgf lines as a string.  It is determined to be sgf lines if
        there is a '\n' character in the data, otherwise it is interpreted as
        a path.

        Note that this scenario just loads the sensor_graph directly into the
        persisted sensor_graph inside the device.  You will still need to
        reset the device for the sensor_graph to enabled and run.

        Args:
            sgf_data (str): Either the path to an sgf file or its contents
                as a string.
        """

        if '\n' not in sgf_data:
            with open(sgf_data, "r") as infile:
                sgf_data = infile.read()

        model = DeviceModel()

        parser = SensorGraphFileParser()
        parser.parse_file(data=sgf_data)

        parser.compile(model)
        opt = SensorGraphOptimizer()
        opt.optimize(parser.sensor_graph, model=model)

        sensor_graph = parser.sensor_graph
        self._logger.info("Loading sensor_graph with %d nodes, %d streamers and %d configs",
                          len(sensor_graph.nodes), len(sensor_graph.streamers), len(sensor_graph.config_database))

        # Directly load the sensor_graph into our persisted storage
        self.sensor_graph.persisted_nodes = sensor_graph.dump_nodes()
        self.sensor_graph.persisted_streamers = sensor_graph.dump_streamers()

        self.sensor_graph.persisted_constants = []
        for stream, value in sorted(sensor_graph.constant_database.items(), key=lambda x: x[0].encode()):
            reading = IOTileReading(stream.encode(), 0, value)
            self.sensor_graph.persisted_constants.append((stream, reading))

        self.sensor_graph.persisted_exists = True

        # Clear all config variables and load in those from this sgf file
        self.config_database.clear()

        for slot in sorted(sensor_graph.config_database, key=lambda x: x.encode()):
            for conf_var, (conf_type, conf_val) in sorted(sensor_graph.config_database[slot].items()):
                self.config_database.add_direct(slot, conf_var, conf_type, conf_val)

        # If we have an app tag and version set program them in
        app_tag = sensor_graph.metadata_database.get('app_tag')
        app_version = sensor_graph.metadata_database.get('app_version')

        if app_tag is not None:
            if app_version is None:
                app_version = "0.0"

            self.app_info = (app_tag, app_version)


def _pack_version(tag, version):
    if tag >= (1 << 20):
        raise ArgumentError("The tag number is too high.  It must fit in 20-bits", max_tag=1 << 20, tag=tag)

    if "." not in version:
        raise ArgumentError("You must pass a version number in X.Y format", version=version)

    major, _, minor = version.partition('.')
    try:
        major = int(major)
        minor = int(minor)
    except ValueError:
        raise ArgumentError("Unable to convert version string into major and minor version numbers", version=version)

    if major < 0 or minor < 0 or major >= (1 << 6) or minor >= (1 << 6):
        raise ArgumentError("Invalid version numbers that must be in the range [0, 63]", major=major, minor=minor, version_string=version)

    version_number = (major << 6) | minor
    combined_tag = (version_number << 20) | tag
    return combined_tag


def _unpack_version(tag_data):
    """Parse a packed version info struct into tag and major.minor version.

    The tag and version are parsed out according to 20 bits for tag and
    6 bits each for major and minor.  The more interesting part is the
    blacklisting performed for tags that are known to be untrustworthy.

    In particular, the following applies to tags.

    - tags < 1024 are reserved for development and have only locally defined
      meaning.  They are not for use in production.
    - tags in [1024, 2048) are production tags but were used inconsistently
      in the early days of Arch and hence cannot be trusted to correspond with
      an actual device model.
    - tags >= 2048 are reserved for supported production device variants.
    - the tag and version 0 (0.0) is reserved for an unknown wildcard that
      does not convey any information except that the tag and version are
      not known.
    """

    tag = tag_data & ((1 << 20) - 1)

    version_data = tag_data >> 20
    major = (version_data >> 6) & ((1 << 6) - 1)
    minor = (version_data >> 0) & ((1 << 6) - 1)

    return (tag, "{}.{}".format(major, minor))



"""Tests of embedded constants and pretty printing."""

from iotile.emulate.constants import stream_name


def test_stream_name_formatting():
    """Make sure stream names are formatted correctly."""

    assert stream_name(0x3800 + 1024) == 'SYSTEM_RESET (0x3C00)'



"""Response objects that allow synchronously waiting for things from the EmulationLoop.

There are two kinds of Response objects: AwaitableResponse that wraps a Future
object for use from a coroutine and CrossThreadResponse that is for waiting
for a response from the event loop by another thread.

Their APIs are identical so that they can be used interchangeably.
"""

import sys
import threading
import asyncio

from iotile.core.exceptions import InternalError, TimeoutExpiredError

class GenericResponse:
    """Base class for asynchronous operation responses."""

    def __init__(self):
        self._result = None
        self._exception = None

    def set_result(self, result):
        """Finish this response and set the result."""

        if self.is_finished():
            raise InternalError("set_result called on finished AsynchronousResponse",
                                result=self._result, exception=self._exception)

        self._result = result
        self.finish()

    def set_exception(self, exc_class, exc_info, exc_stack):
        """Set an exception as the result of this operation.

        Args:
            exc_class (object): The exception type """

        if self.is_finished():
            raise InternalError("set_exception called on finished AsynchronousResponse",
                                result=self._result, exception=self._exception)

        self._exception = (exc_class, exc_info, exc_stack)
        self.finish()

    def capture_exception(self):
        """Capture the current exception context."""

        self.set_exception(*sys.exc_info())

    # Implementation from future.utils.raise_
    def _raise_exception(self):
        if self._exception is None:
            raise InternalError("No exception stored in call to _raise_exception")

        exc_type, value, traceback = self._exception

        if value is not None and isinstance(exc_type, Exception):
            raise TypeError("instance exception may not have a separate value")

        if value is not None:
            exc = exc_type(value)
        else:
            exc = exc_type

        if exc.__traceback__ is not traceback:
            raise exc.with_traceback(traceback)

        raise exc

    def finish(self):
        """Finish this operation without any value."""
        raise NotImplementedError()

    def is_finished(self):
        """Check if this operation is finished.

        Returns:
            bool: True if the operation is finished.
        """
        raise NotImplementedError()


class CrossThreadResponse(GenericResponse):
    """A cross-thread future object that can be waited on.

    This object encapsulates returning a response from the EmulationLoop to a
    caller in a different thread.  It is waitable and can pass either a result
    or an exception, which is raised in the calling thread.
    """

    def __init__(self):
        super(CrossThreadResponse, self).__init__()
        self._finished = threading.Event()
        self._callbacks = []

    def finish(self):
        self._finished.set()

        if len(self._callbacks) > 0:
            for callback in self._callbacks:
                callback(self._exception, self._result)

    def add_callback(self, callback):
        """TEMPORARY METHOD"""
        self._callbacks.append(callback)

    def is_finished(self):
        return self._finished.is_set()

    def wait(self, timeout=None):
        """Wait for this operation to finish.

        You can specify an optional timeout that defaults to no timeout if
        None is passed.  The result of the operation is returned from this
        method. If the operation raised an exception, it is reraised from this
        method.

        Args:
            timeout (float): The maximum number of seconds to wait before timing
                out.
        """

        flag = self._finished.wait(timeout=timeout)
        if flag is False:
            raise TimeoutExpiredError("Timeout waiting for response to event loop operation")

        if self._exception is not None:
            self._raise_exception()

        return self._result


class AwaitableResponse(GenericResponse):
    """Asynchronous response for use inside the event loop."""

    def __init__(self):
        super(AwaitableResponse, self).__init__()
        self._future = asyncio.get_event_loop().create_future()

    def finish(self):
        self._future.set_result(None)

    def is_finished(self):
        return self._future.done()

    async def wait(self, timeout=None):
        """Wait for this operation to finish.

        You can specify an optional timeout that defaults to no timeout if
        None is passed.  The result of the operation is returned from this
        method. If the operation raised an exception, it is reraised from this
        method.

        Args:
            timeout (float): The maximum number of seconds to wait before timing
                out.
        """

        await asyncio.wait_for(self._future, timeout)

        if self._exception is not None:
            self._raise_exception()

        return self._result



"""The clock manager subsystem timestamps data and implements timers.

No other controller subsystem has a concept of device time.  They are all
purely reactive to whatever work items are queued for them.  This is
by design since IOTile devices are meant to be event-driven rather than
polling based whenever possible.

However, there are times when you need to queue operations to be performed
periodically.  It is the clock manager subsystem's job to allow for queuing
these periodic tasks.  Internally, all timers work by just sending a
periodic input to the sensor graph subsystem.  There are 4 timer inputs
available:

- normal tick: generated every 10 seconds in all circumstances.
- fast tick: generated every 1 second when enabled.
- user tick 1: generated at a user configurable interval and easily
  updated dynamically using an RPC.
- user tick 2: generated at a user configurable interval and easily
  updated dynamically using an RPC.

The first two ticks are reserved for internal use and should not be adjusted
by the user, however the last two ticks are designed for easy control in the
field.
"""

from .rpc_declaration import RPCDeclaration

GET_USER_TIMER = RPCDeclaration(0x2014, "H", "LL")
"""Get the current interval in seconds of a user-controllable timer.

This RPC takes an index identifying the timer that you wish to
query and returns the currently configured interval at which that
timer will present its input to the sensor_graph subsystem.

The interval is always returned in seconds.

There are currently 3 possible timers you can query:
  - 0: The fast timer (which should always be set at 1 second)
  - 1: User timer 1.  This is entirely for application use.
  - 2: User timer 2.  This is entirely for application use.

Args:
  - uint16_t: The timer index.  See above for possible values.

Returns:
  - uint32_t: An error code.
  - uint32_t: The configured timer's current period in seconds.  This
    is only valid if there was no error.

Possible Errors:
  - INVALID_ARRAY_KEY: If the index passed does not name a valid timer.  The
    error will be returned from the SENSOR_GRAPH subsystem.
"""

SET_USER_TIMER = RPCDeclaration(0x2015, "LH", "L")
"""Set the current interval in seconds of a user-controllable timer.

This RPC can set the current time interval of a user controllable timer. The
desired time interval is always passed in seconds and a value of 0 will
disable that timer.  A disabled timer does not present an input to the
sensor_graph subsystem.

**This RPC only affects the value of the timer interval until the next time
the controller resets.**

If you want to make the configured change permanent, you need to couple this
call with setting the appropriate config variable as well.

There are currently 3 possible timers you can query:
  - 0: The fast timer (which should always be set at 1 second)
  - 1: User timer 1.  This is entirely for application use.
  - 2: User timer 2.  This is entirely for application use.

The relevant config variables are:
  - fast timer: 0x2000
  - user timer 1: 0x2002
  - user timer 2: 0x2003

Each config variable is a uint32_t that takes the exact same value you would
pass to SET_USER_TIMER.  It has the effect of calling SET_USER_TIMER with that
value on device reset.

Args:
  - uint32_t: The desired timer interval in seconds.  A value of 0 will disable
    the timer.
  - uint16_t: The index of the timer that you wish to configure.

Returns:
  - uint32_t: An error code.

Possible Errors:
  - INVALID_ARRAY_KEY: If the index passed does not name a valid timer.  The
    error will be returned from the SENSOR_GRAPH subsystem.
"""

GET_CURRENT_TIME = RPCDeclaration(0x1001, "", "L")
"""Get the current UTC or uptime of the device.

If the device has a valid UTC time offset set, it will return an encoded UTC
timestamp encoded as the number of seconds since 1/1/2000 00:00Z with the high
bit set to indicate that it is returning UTC time.

Otherwise this RPC will return the number of seconds since the device last
reset with the MSB clear to prevent misinterpretation as UTC time.

Returns:
  - uint32_t: The current UTC or uptime.
"""

GET_CURRENT_UPTIME = RPCDeclaration(0x100f, "", "L")
"""Get the current uptime of the device.

This RPC will always return  the number of seconds since the device last reset
with the MSB clear to prevent misinterpretation as UTC time, even if a valid
UTC time offset has been set.

Returns:
  - uint32_t: The current UTC or uptime.
"""

GET_UTC_TIME_OFFSET = RPCDeclaration(0x100d, "", "LL")
"""Get the currently programmed UTC time offset.

This RPC returns the current seconds offset that is added to device uptime in
order to produce the value returned by the GET_CURRENT_TIME RPC. Typically,
this value is used to adjust the device's time to UTC.  In that case the
second return value will be a 1 indicating that the device is in valid utc
mode.  Otherwise the second result will be a 0.

Returns:
  - uint32_t: The adjustment value applied to uptime to produce the value
    in GET_CURRENT_TIME_RPC.
  - uint32_t: A 1 if the adjustment is declared to produce valid UTC time.
    Otherwise a 0.
"""

SET_UTC_TIME_OFFSET = RPCDeclaration(0x100e, "LL", "L")
"""Set the currently programmed UTC time offset.

This RPC sets an offset in seconds that will be added to the device's current
uptime and returned whenever the GET_CURRENT_TIME RPC is called. Typically,
this value is used to adjust the device's time to UTC.  In that case the
second argument should be a 1 indicating that the device is in valid utc mode.
Otherwise the second argument should be a 0.  There is no real use case for
setting the second argument to 0 except during testing.

**This offset value is volatile and will not survive a device reset.**

If you want to more permanently synchronize the device's time to UTC,
use the SYNCHRONIZE_CLOCK RPC below.

Args:
  - uint32_t: The offset value in seconds to combine with uptime.
  - uint32_t: A 1 if the adjustment is declared to produce valid UTC time.
    Otherwise a 0.

Returns:
  - uint32_t: An error code.  Currently no errors are possible.
"""

SYNCHRONIZE_CLOCK = RPCDeclaration(0x1010, "L", "L")
"""Persistently synchronize the device's clock to UTC.

This RPC is the same as SET_UTC_TIME_OFFSET RPC except the second flag is
hardcoded to 1, indicating that UTC time is always declared and the value
given is passed to an RTC manager to attempt to persist it in hardware so that
it survives a device reset.  This required hardware realtime clock support,
but if present this RPC will allow persistently synchronizing the device's
clock until the hardware rtc is cleared.

Args:
  - uint32_t: The current UTC time in seconds since 1/1/2000 00:00Z.  This
    will have the device's current uptime subtracted from it in order to
    produce the time offset value passed to SET_UTC_TIME_OFFSET.

Returns:
  - uint32_t: An error code.  Currently no errors are possible.
"""



"""Subclass of VirtualDeviceAdapter that supports emulated devices.

This DeviceAdapter provides additional DebugManager functionality such as
saving and loading internal state snapshots and loading test scenarios
into the emulated devices.
"""

import logging
from iotile.core.hw.transport.virtualadapter import VirtualDeviceAdapter
from ..virtual import EmulatedDevice


class EmulatedDeviceAdapter(VirtualDeviceAdapter):
    """DeviceAdapter for connecting to EmulatedDevices.

    This adapter is used the exact same way as VirtualDeviceAdapter, except
    that it only supported EmulatedDevice subclasses and it implements the
    debug interface to allow for dumping and loading the state of the
    emulated device.

    Args:
        port (string): A port description that should be in the form of
            device_name1@<optional_config_json1;device_name2@optional_config_json2
        devices (list of EmulatedDevice): Optional list of specific, precreated emulated
            devices that should be added to the device adapter.
    """

    def __init__(self, port, devices=None):
        super(EmulatedDeviceAdapter, self).__init__(port, devices)
        self._logger = logging.getLogger(__name__)

    @classmethod
    def _validate_device(cls, device):
        """Hook for subclases to ensure that only specific kinds of devices are loaded.

        Returns:
            bool: Whether the virtual device is allowed to load.
        """

        return isinstance(device, EmulatedDevice)

    def debug_async(self, conn_id, cmd_name, cmd_args, progress_callback, callback):
        """Asynchronously complete a named debug command.

        The command name and arguments are passed to the underlying device adapter
        and interpreted there.  If the command is long running, progress_callback
        may be used to provide status updates.  Callback is called when the command
        has finished.

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            cmd_name (string): the name of the debug command we want to invoke
            cmd_args (dict): any arguments that we want to send with this command.
            progress_callback (callable): A function to be called with status on our progress, called as:
                progress_callback(done_count, total_count)
            callback (callable): A callback for when we have finished the debug command, called as:
                callback(connection_id, adapter_id, success, retval, failure_reason)
                'connection_id': the connection id
                'adapter_id': this adapter's id
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
                'retval': A command specific dictionary of return value information
        """

        device = self.connections.get(conn_id)

        if device is None:
            callback(conn_id, self.id, False, None, 'Could not find connection to device.')
            return

        success = True
        reason = None
        retval = None

        try:
            if cmd_name == 'dump_state':
                retval = device.dump_state()
            elif cmd_name == 'restore_state':
                state = cmd_args['snapshot']
                device.restore_state(state)
            elif cmd_name == 'load_scenario':
                scenario = cmd_args['scenario']
                device.load_metascenario(scenario)
            elif cmd_name == 'track_changes':
                if cmd_args['enabled']:
                    device.state_history.enable()
                else:
                    device.state_history.disable()
            elif cmd_name == 'dump_changes':
                outpath = cmd_args['path']
                device.state_history.dump(outpath)
            else:
                success = False
                reason = "Unknown command %s" % cmd_name
        except Exception as exc:  #pylint:disable=broad-except;We need to turn all exceptions into a callback
            self._logger.exception("Error processing debug command %s: args=%s", cmd_name, cmd_args)
            success = False
            reason = "Exception %s occurred during processing" % str(exc)

        callback(conn_id, self.id, success, retval, reason)

    def _open_debug_interface(self, conn_id, callback, connection_string=None):
        """Enable debug interface for this IOTile device

        Args:
            conn_id (int): the unique identifier for the connection
            callback (callback): Callback to be called when this command finishes
                callback(conn_id, adapter_id, success, failure_reason)
        """

        callback(conn_id, self.id, True, None)

    def periodic_callback(self):
        """Periodic callback task.

        This task is called periodically when this device adapter is attached
        to an iotile-gateway and called at specific points in time when it
        is attached to an AdapterStream.

        The primary task that we need to do for emulated devices is to call
        wait_idle() to make sure that they finish any pending background work.

        This is particularly useful for ensuring that reset RPCs happen
        synchronously without needing a delay.
        """

        super(EmulatedDeviceAdapter, self).periodic_callback()

        for dev in self.devices.values():
            dev.wait_idle()



"""Tests to ensure that our coroutine based emulation loop works."""

import pytest
import asyncio
from iotile.emulate.internal import EmulationLoop
from iotile.core.hw.virtual.common_types import AsynchronousRPCResponse
from iotile.core.exceptions import TimeoutExpiredError


def test_basic_eventloop():
    """Make sure basic things work."""

    def _rpc_executor(_address, rpc_id, arg_payload):
        if rpc_id == 0x8000:
            raise ValueError("Error")

        return arg_payload

    loop = EmulationLoop(_rpc_executor)
    loop.start()



    try:
        assert loop.call_rpc_external(8, 0x8001, b'abcd') == b'abcd'

        with pytest.raises(ValueError):
            loop.call_rpc_external(8, 0x8000, b'')

    finally:
        loop.stop()


def test_async_rpc():
    """Make sure we can send an asynchronous RPC."""
    loop = None

    def _rpc_executor(_address, rpc_id, arg_payload):
        if rpc_id == 0x8000:
            raise ValueError("Error")
        elif rpc_id == 0x8002:
            address, rpc_id = loop.get_current_rpc()
            asyncio.get_event_loop().call_soon(loop.finish_async_rpc, address, rpc_id, b'4444')
            raise AsynchronousRPCResponse()

        return arg_payload

    loop = EmulationLoop(_rpc_executor)
    loop.start()

    try:
        assert loop.call_rpc_external(8, 0x8001, b'abcd') == b'abcd'
        assert loop.call_rpc_external(8, 0x8002, b'abcd') == b'4444'
        loop.wait_idle()

    finally:
        loop.stop()


def test_rpc_timeout():
    """Make sure we can timeout an RPC."""

    def _rpc_executor(_address, rpc_id, arg_payload):
        raise AsynchronousRPCResponse()

    loop = EmulationLoop(_rpc_executor)
    loop.start()

    try:
        with pytest.raises(TimeoutExpiredError):
            loop.call_rpc_external(8, 0x8001, b'abcd', timeout=0.001)

        with pytest.raises(TimeoutExpiredError):
            loop.wait_idle(timeout=0.001)

    finally:
        loop.stop()



"""Tests for the EmulatedDeviceAdapter class."""

from __future__ import print_function
import json
from iotile.core.hw import HardwareManager

def save_device_args(tmpdir, filename, data, parent=None):
    """Helper function to dump a config json."""

    if parent is not None:
        data = {parent: data}

    path = tmpdir.join(filename)
    path.write(json.dumps(data), mode="w")
    return str(path)


def test_basic_functionality():
    """Make sure we can connect to a device."""

    with HardwareManager(port='emulated:emulation_demo@#eyJzaW11bGF0ZV90aW1lIjogZmFsc2V9') as hw:
        results = hw.scan()
        assert len(results) == 1
        assert results[0]['uuid'] == 1

        hw.connect(1)
        hw.get(8, basic=True)
        hw.disconnect()


def test_saving_and_loading_state(tmpdir):
    """Make sure we can save and load state."""

    saved = str(tmpdir.join("state.json"))
    with HardwareManager(port='emulated:emulation_demo@#eyJzaW11bGF0ZV90aW1lIjogZmFsc2V9') as hw:
        hw.connect(1)
        debug = hw.debug()

        debug.save_snapshot(saved)
        debug.load_snapshot(saved)


def test_loading_scenario(tmpdir):
    """Make sure we can load a test scenario."""

    scen_file = save_device_args(tmpdir, 'scenario.json', data=[{
        'name': 'loaded_counter',
        'args': {
            'counter': 15
        },
        'tile': 11
    }])

    saved = str(tmpdir.join("state.json"))

    with HardwareManager(port='emulated:emulation_demo@#eyJzaW11bGF0ZV90aW1lIjogZmFsc2V9') as hw:
        hw.connect(1)
        debug = hw.debug()

        debug.open_scenario(scen_file)
        debug.save_snapshot(saved)

    with open(saved, "r") as infile:
        json.load(infile)


def test_saving_changes(tmpdir):
    """Make sure we can track and save changes to a device."""

    scen_file = save_device_args(tmpdir, 'scenario.json', data=[{
        'name': 'loaded_counter',
        'args': {
            'counter': 15,
        },
        'tile': 11
    }])

    change_file = tmpdir.join('out.csv')

    with HardwareManager(port='emulated:emulation_demo@#eyJzaW11bGF0ZV90aW1lIjogZmFsc2V9') as hw:
        hw.connect(1)
        debug = hw.debug()

        debug.track_changes()
        debug.open_scenario(scen_file)
        debug.track_changes(enabled=False)
        debug.save_changes(str(change_file))

    assert change_file.exists()


def test_async_rpc():
    """Make sure we can send an asynchronous rpc."""

    with HardwareManager(port='emulated:emulation_demo@#eyJzaW11bGF0ZV90aW1lIjogZmFsc2V9') as hw:
        hw.connect(1)

        proxy = hw.get(11, basic=True)

        # This RPC is async
        echo, = proxy.rpc_v2(0x8000, "L", "L", 5)
        assert echo == 5

        # This RPC is sync
        echo, = proxy.rpc_v2(0x8001, "L", "L", 6)
        assert echo == 6


def test_racefree_reset():
    """Make sure we can reset at will."""

    with HardwareManager(port='emulated:emulation_demo@#eyJzaW11bGF0ZV90aW1lIjogZmFsc2V9') as hw:
        hw.connect(1)

        proxy = hw.get(8, basic=True)

        for i in range(0, 10):
            print("starting reset %d" % i)
            proxy.reset(wait=0)
            print("finished reset %d" % i)



"""Mixin class to add property change tracking to an emulated device."""

import json
from enum import IntEnum
from iotile.core.exceptions import ArgumentError


class EmulationMixin(object):
    """Mixin class to add property change tracking and state loading to a class.

    This mixin installs a __setattr__ overload that will track whenever a
    write happens to a property contained in _tracked_properties, recording
    the property name and its value.

    It also adds the ability to save and load the state of the emulated object
    so that you can come back to exactly where you were in the future and it
    adds the concept of scenarios, which are preloaded states that can be
    slightly customized with keyword arguments.

    This base class provides standard emulation behavior to both
    EmulatedDevice and EmulatedTile.

    Args:
        address (int): The tile address if we have one, otherwise None
            for device wide changes that do not pertain to a specific
            tile.
        log (EmulationStateLog): The log where we should record our
            changes.
        properties (list of str): Optional list of property names that
            we should track.  If not specified, you can update later
            by adding property named to the `_tracked_properties` set.
    """

    def __init__(self, address, log, properties=None):
        self._emulation_log = log
        self._emulation_address = address

        if properties is None:
            properties = []

        self._tracked_properties = set(properties)
        self._known_scenarios = {}

    def __setattr__(self, name, value):
        if hasattr(self, '_tracked_properties') and name in self._tracked_properties:
            self._emulation_log.track_change(self._emulation_address, name, value)

        super(EmulationMixin, self).__setattr__(name, value)

    def _track_change(self, name, value, formatter=None):
        """Track that a change happened.

        This function is only needed for manually recording changes that are
        not captured by changes to properties of this object that are tracked
        automatically.  Classes that inherit from `emulation_mixin` should
        use this function to record interesting changes in their internal
        state or events that happen.

        The `value` parameter that you pass here should be a native python
        object best representing what the value of the property that changed
        is.  When saved to disk, it will be converted to a string using:
        `str(value)`.  If you do not like the string that would result from
        such a call, you can pass a custom formatter that will be called as
        `formatter(value)` and must return a string.

        Args:
            name (str): The name of the property that changed.
            value (object): The new value of the property.
            formatter (callable): Optional function to convert value to a
                string.  This function will only be called if track_changes()
                is enabled and `name` is on the whitelist for properties that
                should be tracked.  If `formatter` is not passed or is None,
                it will default to `str`
        """

        self._emulation_log.track_change(self._emulation_address, name, value, formatter)

    def dump_state(self):
        """Dump the current state of this emulated object as a dictionary.

        Returns:
            dict: The current state of the object that could be passed to load_state.
        """

        raise NotImplementedError("All subclasses must override dump_state")

    def restore_state(self, state):
        """Restore the current state of this emulated object.

        Args:
            state (dict): A previously dumped state produced by dump_state.
        """

        raise NotImplementedError("All subclasses must override restore_state")

    def save_state(self, out_path):
        """Save the current state of this emulated object to a file.

        Args:
            out_path (str): The path to save the dumped state of this emulated
                object.
        """

        state = self.dump_state()

        # Remove all IntEnums from state since they cannot be json-serialized on python 2.7
        # See https://bitbucket.org/stoneleaf/enum34/issues/17/difference-between-enum34-and-enum-json
        state = _clean_intenum(state)

        with open(out_path, "w") as outfile:
            json.dump(state, outfile, indent=4)

    def load_state(self, in_path):
        """Load the current state of this emulated object from a file.

        The file should have been produced by a previous call to save_state.

        Args:
            in_path (str): The path to the saved state dump that you wish
                to load.
        """

        with open(in_path, "r") as infile:
            state = json.load(infile)

        self.restore_state(state)

    def load_scenario(self, scenario_name, **kwargs):
        """Load a scenario into the emulated object.

        Scenarios are specific states of an an object that can be customized
        with keyword parameters.  Typical examples are:

          - data logger with full storage
          - device with low battery indication on

        Args:
            scenario_name (str): The name of the scenario that we wish to
                load.
            **kwargs: Any arguments that should be passed to configure
                the scenario.  These arguments will be passed directly
                to the scenario handler.
        """

        scenario = self._known_scenarios.get(scenario_name)
        if scenario is None:
            raise ArgumentError("Unknown scenario %s" % scenario_name, known_scenarios=list(self._known_scenarios))

        scenario(**kwargs)

    def register_scenario(self, scenario_name, handler):
        """Register a scenario handler for this object.

        Scenario handlers are callable functions with no positional arguments
        that can be called by name with the load_scenario function and should
        prepare the emulated object into a known state.  The purpose of a
        scenario is to make it easy to get a device into a specific state for
        testing purposes that may otherwise be difficult or time consuming to
        prepare on the physical, non-emulated device.

        Args:
            scenario_name (str): The name of this scenario that can be passed to
                load_scenario later in order to invoke the scenario.
            handler (callable): A callable function that takes no positional
                arguments and can prepare this object into the given scenario
                state.  It may take required or optional keyword arguments that
                may be passed to `load_scenario` if needed.
        """

        if scenario_name in self._known_scenarios:
            raise ArgumentError("Attempted to add the same scenario name twice", scenario_name=scenario_name,
                                previous_handler=self._known_scenarios[scenario_name])

        self._known_scenarios[scenario_name] = handler


def _clean_intenum(obj):
    """Remove all IntEnum classes from a map."""

    if isinstance(obj, dict):
        for key, value in obj.items():
            if isinstance(value, IntEnum):
                obj[key] = value.value
            elif isinstance(value, (dict, list)):
                obj[key] = _clean_intenum(value)
    elif isinstance(obj, list):
        for i, value in enumerate(obj):
            if isinstance(value, IntEnum):
                obj[i] = value.value
            elif isinstance(value, (dict, list)):
                obj[i] = _clean_intenum(value)

    return obj



"""Base class for all emulated non-controller tiles."""

from iotile.core.hw.virtual import tile_rpc
from .emulated_tile import EmulatedTile
from ..constants import rpcs, RunLevel


class EmulatedPeripheralTile(EmulatedTile):
    """Base class for all emulated non-controller tiles.

    Non-controller (also called peripheral) tiles wait to start until they are
    signaled to do so from a controller tile.  They also expect configuration
    information to be streamed to them in the form of config variables by the
    controller before it calls the start_application RPC that causes the
    peripheral tile to latch in the state of its config variables and start
    operation.

    Args:
        address (int): The address of this tile in the VirtualIOTIleDevice
            that contains it
        device (TileBasedVirtualDevice): Device on which this tile is running.
            This parameter is not optional on EmulatedTiles.
    """

    __NO_EXTENSION__ = True

    def __init__(self, address, device):
        EmulatedTile.__init__(self, address, device)

        self._registered = device.emulator.create_event()
        self._start_received = device.emulator.create_event()
        self._hosted_app_running = device.emulator.create_event()

        self.debug_mode = False
        self.run_level = None

    async def _application_main(self):
        """The placeholder for the main application loop.

        This method should be overwritten by subclasses of
        EmulatedPeripheralTile with their background main loop task.  This
        coroutine is equivalent to the main() C function in a physical tile's
        firmware.

        This coroutine should loop forever.  It will not be restarted if it
        returns unless reset() is called on the tile.

        If you override this method, you **must set self.initialized()** once
        you have initialized any required state, otherwise the emulation will
        hang waiting for your tile to indicate that it has been initialized.
        """

        self.initialized.set()

    async def _reset_vector(self):
        """Main background task for the tile executive.

        The tile executive is in charge registering the tile with the
        controller and then handing control over to the tile's application
        firmware after proper configuration values have been received.
        """

        self._logger.info("Tile %s at address %d is starting from reset", self.name, self.address)

        try:
            address, run_level, debug = await self._device.emulator.await_rpc(8, rpcs.REGISTER_TILE, *self._registration_tuple())
        except:
            self._logger.exception("Error registering tile: address=%d, name=%s", self.address, self.name)
            raise

        self.debug_mode = bool(debug)
        self.run_level = run_level
        self._logger.info("Tile at address %d registered itself, received address=%d, runlevel=%d and debug=%d", self.address, address, run_level, debug)

        self._registered.set()

        # If we are in safe mode we do not run the main application
        # loop.
        if run_level == RunLevel.SAFE_MODE:
            self.initialized.set()
            return

        if run_level == RunLevel.START_ON_COMMAND:
            await self._start_received.wait()

        self._hosted_app_running.set()
        await self._application_main()

    def _registration_tuple(self):
        return (self.hardware_type, self.api_version[0], self.api_version[1], self.name,
                self.firmware_version[0], self.firmware_version[1], self.firmware_version[2],
                self.executive_version[0], self.executive_version[1], self.executive_version[2],
                self.address - 10, 0)

    def _handle_reset(self):
        """Reset this tile.

        This process needs to trigger the peripheral tile to reregister itself
        with the controller and get new configuration variables.  It also
        needs to clear app_running.
        """

        self._registered.clear()
        self._start_received.clear()
        self._hosted_app_running.clear()

        super(EmulatedPeripheralTile, self)._handle_reset()

    def dump_state(self):
        """Dump the current state of this emulated tile as a dictionary.

        This function just dumps the status of the config variables.  It is
        designed to be called in a chained fashion to serialize the complete
        state of a tile subclass.

        Returns:
            dict: The current state of the object that could be passed to load_state.
        """

        state = super(EmulatedPeripheralTile, self).dump_state()
        state['app_started'] = self._hosted_app_running.is_set()
        state['debug_mode'] = self.debug_mode
        state['run_level'] = self.run_level

        return state

    def restore_state(self, state):
        """Restore the current state of this emulated object.

        Args:
            state (dict): A previously dumped state produced by dump_state.
        """

        super(EmulatedPeripheralTile, self).restore_state(state)

        self.debug_mode = state.get('debug_mode', False)
        self.run_level = state.get('run_level', None)

        if state.get('app_started', False):
            self._hosted_app_running.set()

    @tile_rpc(*rpcs.START_APPLICATION)
    def start_application(self):
        """Latch any configuration variables and start the application."""

        self._start_received.set()



"""Helper class for dispatching rpcs inside the emulation loop."""

import asyncio
import logging
import inspect
from iotile.core.exceptions import InternalError, ArgumentError
from iotile.core.hw.virtual.common_types import AsynchronousRPCResponse

class RPCQueue:
    """Coroutine based RPC dispatcher.

    This class wraps a coroutine draining an asyncio.Queue to dispatch RPCS.
    It keeps track of RPCs that are in progress and allows joining the queue
    to wait for an idle moment when no RPCs are pending or in progress.

    This class is primarily used to simplify the implementation of
    EmulationLoop, which creates an RPCQueue and interacts with it.

    The primary mode of interaction with this class is to call:

    - put_rpc(address, rpc_id, payload, response)
    - put_task(func, args, response)

    Those methods will add work to the RPC queue, which will be performed
    asynchronously by the background dispatcher task.  When a work item is
    finished, the response object passed in will be notified with the result.

    Args:
        loop (asyncio.EventLoop): The event loop this queue should be attached
            to.
        rpc_handler (callable): The actual handler function that we should use
            to run each RPC and get the result.
    """

    def __init__(self, loop, rpc_handler):
        self._rpc_handler = rpc_handler
        self._loop = loop
        self._rpc_queue = asyncio.Queue(loop=loop)
        self._current_rpc = None
        self._rpc_task = None
        self._pending_rpcs = {}
        self._logger = logging.getLogger(__name__)

    def put_task(self, func, args, response):
        """Place a task onto the RPC queue.

        This temporary functionality will go away but it lets you run a
        task synchronously with RPC dispatch by placing it onto the
        RCP queue.

        Args:
            func (callable): The function to execute
            args (iterable): The function arguments
            response (GenericResponse): The response object to signal the
                result on.
        """

        self._rpc_queue.put_nowait((func, args, response))

    def put_rpc(self, address, rpc_id, arg_payload, response):
        """Place an RPC onto the RPC queue.

        The rpc will be dispatched asynchronously by the background dispatch
        task.  This method must be called from the event loop.  This method
        does not block.

        Args:
            address (int): The address of the tile with the RPC
            rpc_id (int): The id of the rpc you want to call
            arg_payload (bytes): The RPC payload
            respones (GenericResponse): The object to use to signal the result.
        """

        self._rpc_queue.put_nowait((address, rpc_id, arg_payload, response))

    def join(self):
        """Wait for the RPC queue to be empty.

        Returns:
            awaitable
        """

        return self._rpc_queue.join()

    def get_current_rpc(self):
        """Get the currently running RPC for asynchronous responses.

        Returns the address and rpc_id of the RPC that is currently being
        dispatched. This information can be saved and passed later to
        finish_async_rpc() to send a response to an asynchronous rpc.

        Returns:
            (address, rpc_id): A tuple with the currently running RPC.
        """

        if self._current_rpc is None:
            raise InternalError("There is no current RPC running")

        return self._current_rpc

    def finish_async_rpc(self, address, rpc_id, response):
        """Finish a previous asynchronous RPC.

        This method should be called by a peripheral tile that previously
        had an RPC called on it and chose to response asynchronously by
        raising ``AsynchronousRPCResponse`` in the RPC handler itself.

        The response passed to this function will be returned to the caller
        as if the RPC had returned it immediately.

        This method must only ever be called from a coroutine inside the
        emulation loop that is handling background work on behalf of a tile.

        Args:
            address (int): The tile address the RPC was called on.
            rpc_id (int): The ID of the RPC that was called.
            response (bytes): The bytes that should be returned to
                the caller of the RPC.
        """

        pending = self._pending_rpcs.get(address)

        if pending is None:
            raise ArgumentError("No asynchronously RPC currently in progress on tile %d" % address)

        responder = pending.get(rpc_id)
        if responder is None:
            raise ArgumentError("RPC %04X is not running asynchronous on tile %d" % (rpc_id, address))

        del pending[rpc_id]

        responder.set_result(response)
        self._rpc_queue.task_done()

    def start(self):
        """Start this task from the event loop thread."""

        self._rpc_task = self._loop.create_task(self._rpc_dispatch_task())

    async def stop(self):
        """Stop the rpc queue from inside the event loop."""

        if self._rpc_task is not None:
            self._rpc_task.cancel()

        try:
            await self._rpc_task
        except asyncio.CancelledError:
            pass

        self._rpc_task = None

    async def _rpc_dispatch_task(self):
        self._logger.debug("Starting RPC dispatch task")

        while True:
            try:
                obj = await self._rpc_queue.get()

                if len(obj) == 3:
                    address = None
                    rpc_id = None
                    func, args, response = obj
                    result = func(*args)
                else:
                    address, rpc_id, arg_payload, response = obj

                    self._current_rpc = (address, rpc_id)
                    result = self._rpc_handler(address, rpc_id, arg_payload)

                    if inspect.iscoroutine(result):
                        result = await result

                    self._current_rpc = None

                response.set_result(result)
                self._rpc_queue.task_done()
            except AsynchronousRPCResponse:
                self._queue_async_rpc(address, rpc_id, response)
            except asyncio.CancelledError:
                raise
            except:  #pylint:disable=bare-except;We want to send all exceptions to the caller.
                response.capture_exception()
                self._rpc_queue.task_done()

    def _queue_async_rpc(self, address, rpc_id, response):
        if address not in self._pending_rpcs:
            self._pending_rpcs[address] = {}

        self._pending_rpcs[address][rpc_id] = response

        self._logger.debug("Queued asynchronous rpc on tile %d, rpc_id: 0x%04X", address, rpc_id)



"""Mixin for persistent database of config variables to send to tiles on reset."""

import base64
import struct
from iotile.core.exceptions import ArgumentError, DataError
from iotile.core.hw.virtual import tile_rpc
from iotile.sg import SlotIdentifier
from ...virtual import SerializableState
from ...constants import Error, ConfigDatabaseError, rpcs


class ConfigEntry(object):
    """Data class representing a single config variable entry."""

    CONTROL_SIZE = 16

    def __init__(self, selector, var_id, data, valid=True):
        self.target = selector
        self.var_id = var_id
        self.data = data
        self.valid = valid

    def control_space(self):
        """Return how much control space this entry uses."""
        return self.CONTROL_SIZE

    def data_space(self):
        """Return how much data space this entry uses.

        The variable target is stored prepended to the data so even if there
        is no actual data, there are always two bytes used.

        Returns:
            int
        """

        return len(self.data)

    def dump(self):
        """Serialize this object."""

        return {
            'target': str(self.target),
            'data': base64.b64encode(self.data).decode('utf-8'),
            'var_id': self.var_id,
            'valid': self.valid
        }

    def generate_rpcs(self, address):
        """Generate the RPCs needed to stream this config variable to a tile.

        Args:
            address (int): The address of the tile that we should stream to.

        Returns:
            list of tuples: A list of argument tuples for each RPC.

            These tuples can be passed to EmulatedDevice.rpc to actually make
            the RPCs.
        """

        rpc_list = []

        for offset in range(2, len(self.data), 16):
            rpc = (address, rpcs.SET_CONFIG_VARIABLE, self.var_id, offset - 2, self.data[offset:offset + 16])
            rpc_list.append(rpc)

        return rpc_list

    @classmethod
    def Restore(cls, state):
        """Unserialize this object."""

        target = SlotIdentifier.FromString(state.get('target'))
        data = base64.b64decode(state.get('data'))
        var_id = state.get('var_id')
        valid = state.get('valid')

        return ConfigEntry(target, var_id, data, valid)


class ConfigDatabase(SerializableState):
    """Emulated state of the config database.

    Args:
        control_size (int): The number of bytes reserved for storing control
            and targeting information.
        data_size (int): The number of bytes reseved for storing config data.
    """

    ENTRY_MAGIC = 0xcaca

    def __init__(self, control_size, data_size):
        super(ConfigDatabase, self).__init__()

        self.control_size = control_size
        self.data_size = data_size
        self.entries = []
        self.in_progress = None
        self.data_index = 0

        self.mark_complex('entries', lambda entries: [x.dump() for x in entries], lambda entries: [ConfigEntry.Restore(x) for x in entries])

    def max_entries(self):
        """Calculate the maximum number of config variables that we can store."""

        return self.control_size // ConfigEntry.CONTROL_SIZE - 1

    def compact(self):
        """Remove all invalid config entries."""

        saved_length = 0
        to_remove = []
        for i, entry in enumerate(self.entries):
            if not entry.valid:
                to_remove.append(i)
                saved_length += entry.data_space()

        for i in reversed(to_remove):
            del self.entries[i]

        self.data_index -= saved_length

    def clear(self):
        """Clear all config variables stored in this config database."""

        self.entries = []
        self.in_progress = None
        self.data_index = 0

    def start_entry(self, target, var_id):
        """Begin a new config database entry.

        If there is a current entry in progress, it is aborted but the
        data was already committed to persistent storage so that space
        is wasted.

        Args:
            target (SlotIdentifer): The target slot for this config variable.
            var_id (int): The config variable ID

        Returns:
            int: An error code from the global Errors enum.
        """

        self.in_progress = ConfigEntry(target, var_id, b'')

        if self.data_size - self.data_index < self.in_progress.data_space():
            return Error.DESTINATION_BUFFER_TOO_SMALL

        self.in_progress.data += struct.pack("<H", var_id)
        self.data_index += self.in_progress.data_space()

        return Error.NO_ERROR

    def add_data(self, data):
        """Add data to the currently in progress entry.

        Args:
            data (bytes): The data that we want to add.

        Returns:
            int: An error code
        """

        if self.data_size - self.data_index < len(data):
            return Error.DESTINATION_BUFFER_TOO_SMALL

        if self.in_progress is not None:
            self.in_progress.data += data

        return Error.NO_ERROR

    def end_entry(self):
        """Finish a previously started config database entry.

        This commits the currently in progress entry.  The expected flow is
        that start_entry() is called followed by 1 or more calls to add_data()
        followed by a single call to end_entry().

        Returns:
            int: An error code
        """

        # Matching current firmware behavior
        if self.in_progress is None:
            return Error.NO_ERROR

        # Make sure there was actually data stored
        if self.in_progress.data_space() == 2:
            return Error.INPUT_BUFFER_WRONG_SIZE

        # Invalidate all previous copies of this config variable so we
        # can properly compact.
        for entry in self.entries:
            if entry.target == self.in_progress.target and entry.var_id == self.in_progress.var_id:
                entry.valid = False

        self.entries.append(self.in_progress)
        self.data_index += self.in_progress.data_space() - 2  # Add in the rest of the entry size (we added two bytes at start_entry())
        self.in_progress = None

        return Error.NO_ERROR

    def stream_matching(self, address, name):
        """Return the RPCs needed to stream matching config variables to the given tile.

        This function will return a list of tuples suitable for passing to
        EmulatedDevice.deferred_rpc.

        Args:
            address (int): The address of the tile that we wish to stream to
            name (str or bytes): The 6 character name of the target tile.

        Returns:
            list of tuple: The list of RPCs to send to stream these variables to a tile.
        """

        matching = [x for x in self.entries if x.valid and x.target.matches(address, name)]

        rpc_list = []
        for var in matching:
            rpc_list.extend(var.generate_rpcs(address))

        return rpc_list

    def add_direct(self, target, var_id, var_type, data):
        """Directly add a config variable.

        This method is meant to be called from emulation scenarios that
        want to directly set config database entries from python.

        Args:
            target (SlotIdentifer): The target slot for this config variable.
            var_id (int): The config variable ID
            var_type (str): The config variable type
            data (bytes or int or str): The data that will be encoded according
                to var_type.
        """

        data = struct.pack("<H", var_id) + _convert_to_bytes(var_type, data)

        if self.data_size - self.data_index < len(data):
            raise DataError("Not enough space for data in new conig entry", needed_space=len(data), actual_space=(self.data_size - self.data_index))

        new_entry = ConfigEntry(target, var_id, data)

        for entry in self.entries:
            if entry.target == new_entry.target and entry.var_id == new_entry.var_id:
                entry.valid = False

        self.entries.append(new_entry)
        self.data_index += new_entry.data_space()


class ConfigDatabaseMixin(object):
    """Persistent data of config variables.

    These variables are targeted at at a tile and are automatically streamed
    to that tile when it registers.

    The config variables are stored as binary blobs in flash memory with a
    fixed area of flash reserved for them.

    The config database is implemented

    Args:
        control_size (int): The number of bytes reserved for storing control
            and targeting information.
        data_size (int): The number of bytes reseved for storing config data.
    """

    def __init__(self, control_size, data_size):
        self.config_database = ConfigDatabase(control_size, data_size)

    @tile_rpc(*rpcs.START_CONFIG_VAR_ENTRY)
    def start_config_var_entry(self, var_id, encoded_selector):
        """Start a new config variable entry."""

        selector = SlotIdentifier.FromEncoded(encoded_selector)

        err = self.config_database.start_entry(selector, var_id)
        return [err]

    @tile_rpc(*rpcs.CONTINUE_CONFIG_VAR_ENTRY)
    def continue_config_var_entry(self, data):
        """Push data to the current config variable entry."""

        err = self.config_database.add_data(data)
        return [err]

    @tile_rpc(*rpcs.END_CONFIG_VAR_ENTRY)
    def end_config_var_entry(self):
        """Finish the currently in progress config variable entry."""

        err = self.config_database.end_entry()
        return [err]

    @tile_rpc(*rpcs.GET_CONFIG_VAR_ENTRY)
    def get_config_var_entry(self, index):
        """Get the metadata from the selected config variable entry."""

        if index == 0 or index > len(self.config_database.entries):
            return [Error.INVALID_ARRAY_KEY, 0, 0, 0, b'\0'*8, 0, 0]

        entry = self.config_database.entries[index - 1]
        if not entry.valid:
            return [ConfigDatabaseError.OBSOLETE_ENTRY, 0, 0, 0, b'\0'*8, 0, 0]

        offset = sum(x.data_space() for x in self.config_database.entries[:index - 1])
        return [Error.NO_ERROR, self.config_database.ENTRY_MAGIC, offset, entry.data_space(), entry.target.encode(), 0xFF, 0]

    @tile_rpc(*rpcs.COUNT_CONFIG_VAR_ENTRIES)
    def count_config_var_entries(self):
        """Count how many config variables have been saved."""

        return [len(self.config_database.entries)]

    @tile_rpc(*rpcs.CLEAR_CONFIG_VAR_ENTRIES)
    def clear_config_var_entries(self):
        """Clear all currently stored config variables."""

        self.config_database.clear()
        return [Error.NO_ERROR]

    @tile_rpc(*rpcs.GET_CONFIG_VAR_ENTRY_DATA)
    def get_config_var_data(self, index, offset):
        """Get a chunk of data for a config variable."""

        if index == 0 or index > len(self.config_database.entries):
            return [Error.INVALID_ARRAY_KEY, b'']

        entry = self.config_database.entries[index - 1]
        if not entry.valid:
            return [ConfigDatabaseError.OBSOLETE_ENTRY, b'']

        if offset >= len(entry.data):
            return [Error.INVALID_ARRAY_KEY, b'']

        data_chunk = entry.data[offset:offset + 16]
        return [Error.NO_ERROR, data_chunk]

    @tile_rpc(*rpcs.INVALIDATE_CONFIG_VAR_ENTRY)
    def invalidate_config_var_entry(self, index):
        """Mark a config variable as invalid."""

        if index == 0 or index > len(self.config_database.entries):
            return [Error.INVALID_ARRAY_KEY, b'']

        entry = self.config_database.entries[index - 1]
        if not entry.valid:
            return [ConfigDatabaseError.OBSOLETE_ENTRY, b'']

        entry.valid = False
        return [Error.NO_ERROR]

    @tile_rpc(*rpcs.COMPACT_CONFIG_DATABASE)
    def compact_config_database(self):
        """Compact the config database, removing any invalid entries."""

        self.config_database.compact()
        return [Error.NO_ERROR]

    @tile_rpc(*rpcs.GET_CONFIG_DATABASE_INFO)
    def get_config_database_info(self):
        """Get memory usage and space statistics on the config database."""

        max_size = self.config_database.data_size
        max_entries = self.config_database.max_entries()
        used_size = self.config_database.data_index
        used_entries = len(self.config_database.entries)
        invalid_size = sum(x.data_space() for x in self.config_database.entries if not x.valid)
        invalid_entries = sum(1 for x in self.config_database.entries if not x.valid)

        return [max_size, used_size, invalid_size, used_entries, invalid_entries, max_entries, 0]


def _convert_to_bytes(type_name, value):
    """Convert a typed value to a binary array"""

    int_types = {'uint8_t': 'B', 'int8_t': 'b', 'uint16_t': 'H', 'int16_t': 'h', 'uint32_t': 'L', 'int32_t': 'l'}

    type_name = type_name.lower()

    if type_name not in int_types and type_name not in ['string', 'binary']:
        raise ArgumentError('Type must be a known integer type, integer type array, string', known_integers=int_types.keys(), actual_type=type_name)

    if type_name == 'string':
        #value should be passed as a string
        bytevalue = bytes(value)
    elif type_name == 'binary':
        bytevalue = bytes(value)
    else:
        bytevalue = struct.pack("<%s" % int_types[type_name], value)

    return bytevalue



"""Package setup file for iotile-emulate."""

from setuptools import setup, find_packages
import version

setup(
    name="iotile-emulate",
    packages=find_packages(exclude=("test",)),
    version=version.version,
    license="LGPLv3",
    description="IOTile Device Emulation",
    install_requires=[
        "iotile-core>=4.0.0",
        "iotile-sensorgraph>=0.8.1"
    ],
    python_requires=">=3.5,<4",
    entry_points={'iotile.virtual_device': ['reference_1_0 = iotile.emulate.demo:DemoReferenceDevice',
                                            'emulation_demo = iotile.emulate.demo:DemoEmulatedDevice'],
                  'iotile.proxy': ['emudmo = iotile.emulate.demo:DemoTileProxy'],
                  'iotile.device_adapter': ['emulated = iotile.emulate.transport:EmulatedDeviceAdapter'],
                  'iotile.emulated_tile': ['refcon_1 = iotile.emulate.reference:ReferenceController']},
    author="Arch",
    author_email="info@arch-iot.com",
    url="https://github.com/iotile/coretools/iotilesensorgraph",
    keywords=["iotile", "arch", "embedded", "hardware"],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
    long_description="""\
iotile-emulate
--------------

A package that support emulating physical iotile devices based on python based emulation models.

See https://www.archsys.io.
"""
)
